/*
 * CORBASkeletonGen.java
 *
 * Created on 13. duben 2002, 21:08
 */

package SOFA.Connector.EEG.EEM.Elements;

import SOFA.SOFAnode.Made.TIR.*;
import SOFA.SOFAnode.Made.TIR.Access.TIRAccessMethods;
import SOFA.Connector.EEG.CodeWriter.*;
import SOFA.Connector.EEG.EEM.Actions.*;

/**
 *
 * @author  ghort
 * @version 
 */
public class CORBASkeletonGen extends JImplWriterDirectorBase {

	/** Creates new CORBASkeletonGen */
	public CORBASkeletonGen(String provIfaceCDL, SOFA.Connector.Property[] params) {
		super(provIfaceCDL,params);
	}

	public String getExtends() {
		String corbaIface=SOFA.Connector.Property.findFirst(params,"provIfaceMapped");
		return corbaIface+"POA";
	}

	public String[] getImplements() {
		return new String[] {"SOFA.Connector.EEG.Types.Skeleton"};
	}

	public boolean defaultImplements() {
		return false;
	}
	
	public String getClassBody() {
		return managerTieImpl() + "\n" + linkImpl() + "\n" +
			"\tprotected String corbaIOR;\n"+
			"\tprotected org.omg.CORBA.Object corbaObject;\n\n"+
			"\tpublic CORBASkeleton() throws SOFA.Connector.ConnectorException {\n"+
			"\t\tsuper();\n"+
			"\t\ttry {\n"+
			"\t\t\tcorbaObject=SOFA.Util.CORBAAccess.activate_object(this);\n"+
			"\t\t\tcorbaIOR=SOFA.Util.CORBAAccess.object_to_string(corbaObject);\n"+
			"\t\t} catch (Exception e) {\n"+
			"\t\t\tthrow new SOFA.Connector.ConnectorException(\"Can't bind CORBASkeleton to '\"+corbaIOR+\"'.\",e);\n"+
			"\t\t}\n"+
			"\t}\n\n"+
			"\tpublic SOFA.Connector.TaggedProfile[] getProfiles() {\n"+
			"\t\treturn new SOFA.Connector.TaggedProfile[] { new SOFA.Connector.ECG.Profiles.CORBAProfile(corbaIOR) };\n"+
			"\t}\n";
	}

	public String getMethodDef(OperationDef mt) throws JWriterDirectorException {
		return JIfaceCNCORBA.getMethodDef(mt);
	}

	public String getMethodBody(OperationDef method) throws JWriterDirectorException {
		try {
			String ret="";

			int i;
			ParamDescr[] params=method.params();

			for (i=0;i<params.length;i++) {
				int mode=params[i].mode().value();
				if (mode==ParamMode.PARAM_INOUT || mode==ParamMode.PARAM_OUT) {
					String holderName=CDL2JavaMapping.getTypeName(params[i].type(),mode);
					ret+="\t\t"+holderName+" "+params[i].name()+"=new "+holderName+"();  "+params[i].name()+".value=_corbaholder_"+params[i].name()+".value;\n";
				}
			}

			ret+="\t\t"+declareRet(method,"ret")+"\n"+
				"\t\t"+assignRet(method,"ret")+targetCall(method,JIfaceCN.targetCallParams(method))+";\n";

			for (i=0;i<params.length;i++) {
				int mode=params[i].mode().value();
				if (mode==ParamMode.PARAM_INOUT || mode==ParamMode.PARAM_OUT) {
					ret+="\t\t_corbaholder_"+params[i].name()+".value="+params[i].name()+".value;\n";
				}
			}
			
			ret+="\t\t"+returnRet(method,"ret")+"\n";

			return ret;

		} catch (java.rmi.RemoteException e) {
			throw new JWriterDirectorException("Cannot access TIR.",e);
		} catch (TIRExceptLock e) {
			throw new JWriterDirectorException("Cannot access TIR.",e);
		} catch (CodeWriterException e) {
			throw new JWriterDirectorException("Cannot map operation to Java.",e);
		}
	}
}
