/* $Id: CompProtOper.java,v 2.1 2002/04/19 12:25:51 hnetynka Exp $ */
package SOFA.SOFAnode.Made.CDL;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.RemoteException;

abstract class CompProtOper {
  public int type;
  public CompProtOper(int t) { type = t; }
  abstract public String toText();

  public static ProtocolOperDef toNormalIface(CompProtOper p, ProtocolDef newprot) throws CDLExceptRemote, CDLExceptToNormal {
    try {
      String[] s =  new String[1];
      ProtocolOperDef ret = null;
      switch (p.type) {
      case CompProtKind.p_operand:
        s[0]=((CompProtOperand) p).operand;
        ret = newprot.create_operand(ProtocolMode.pm_none, s);
        break;
      case CompProtKind.p_nulloperand:
        ret = newprot.create_nulloperand();
        break;
      case CompProtKind.p_metaoperand:
        ret = newprot.create_metaoperand(((CompProtMetaOperand) p).name);
        break;
      case CompProtKind.p_repetition:
        ret = newprot.create_unoperation(ProtocolOperKind.pok_repetition, toNormalIface(((CompProtUnOper) p).operand, newprot));
        break;
      case CompProtKind.p_andparallel:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_andparallel, toNormalIface(((CompProtBinOper) p).operand1, newprot), toNormalIface(((CompProtBinOper) p).operand2, newprot));
        break;
      case CompProtKind.p_orparallel:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_orparallel, toNormalIface(((CompProtBinOper) p).operand1, newprot), toNormalIface(((CompProtBinOper) p).operand2, newprot));
        break;
      case CompProtKind.p_alternative:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_alternative, toNormalIface(((CompProtBinOper) p).operand1, newprot), toNormalIface(((CompProtBinOper) p).operand2, newprot));
        break;
      case CompProtKind.p_sequence:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_sequence, toNormalIface(((CompProtBinOper) p).operand1, newprot), toNormalIface(((CompProtBinOper) p).operand2, newprot));
        break;
      case CompProtKind.p_restriction:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_restriction, toNormalIface(((CompProtBinOper) p).operand1, newprot), toNormalIface(((CompProtBinOper) p).operand2, newprot));
        break;
      default: break;
      }
      if (ret==null)
        throw new CDLExceptToNormal("Can't create protocol");
      return ret;
    } catch (RemoteException e) {
      throw new CDLExceptRemote("Remote exception occured: "+e.getMessage());
    } catch (TIRExceptCreate e) {
      throw new CDLExceptToNormal("Can't create protocol: "+e.getMessage());
    }
  }

  public static ProtocolOperDef toNormalFrame(CompProtOper p, ProtocolDef newprot) throws CDLExceptRemote, CDLExceptToNormal {
    try {
      String[] s =  new String[2];
      ProtocolOperDef ret = null;
      switch (p.type) {
      case CompProtKind.p_operand:
        s[0]=((CompProtOperand) p).ifacename;
        s[1]=((CompProtOperand) p).operand;
        ret = newprot.create_operand(((CompProtOperand) p).mode, s);
        break;
      case CompProtKind.p_nulloperand:
        ret = newprot.create_nulloperand();
        break;
      case CompProtKind.p_metaoperand:
        ret = newprot.create_metaoperand(((CompProtMetaOperand) p).name);
        break;
      case CompProtKind.p_nestedoperand:
        s[0]=((CompProtNestedOperand) p).ifacename;
        s[1]=((CompProtNestedOperand) p).operand;
        ret = newprot.create_nestedoperand(((CompProtNestedOperand) p).mode, s, toNormalFrame(((CompProtNestedOperand) p).nestOperand, newprot));
        break;
      case CompProtKind.p_repetition:
        ret = newprot.create_unoperation(ProtocolOperKind.pok_repetition, toNormalFrame(((CompProtUnOper) p).operand, newprot));
        break;
      case CompProtKind.p_andparallel:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_andparallel, toNormalFrame(((CompProtBinOper) p).operand1, newprot), toNormalFrame(((CompProtBinOper) p).operand2, newprot));
        break;
      case CompProtKind.p_orparallel:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_orparallel, toNormalFrame(((CompProtBinOper) p).operand1, newprot), toNormalFrame(((CompProtBinOper) p).operand2, newprot));
        break;
      case CompProtKind.p_alternative:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_alternative, toNormalFrame(((CompProtBinOper) p).operand1, newprot), toNormalFrame(((CompProtBinOper) p).operand2, newprot));
        break;
      case CompProtKind.p_sequence:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_sequence, toNormalFrame(((CompProtBinOper) p).operand1, newprot), toNormalFrame(((CompProtBinOper) p).operand2, newprot));
        break;
      case CompProtKind.p_restriction:
        ret = newprot.create_binoperation(ProtocolOperKind.pok_restriction, toNormalFrame(((CompProtBinOper) p).operand1, newprot), toNormalFrame(((CompProtBinOper) p).operand2, newprot));
        break;
      default: break;
      }
      if (ret==null)
        throw new CDLExceptToNormal("Can't create protocol");
      return ret;
    } catch (RemoteException e) {
      throw new CDLExceptRemote("Remote exception occured: "+e.getMessage());
    } catch (TIRExceptCreate e) {
      throw new CDLExceptToNormal("Can't create protocol: "+e.getMessage());
    }
  }
}
