/* $Id: CompStruct.java,v 2.1 2002/04/17 13:41:40 hnetynka Exp $ */
package SOFA.SOFAnode.Made.CDL;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.RemoteException;

class CompStruct extends CompContainedContainerIm implements CompType {
  public CompStruct(ID id, CompContainer in, CompRepository inRep) {
    super(ObjectsKind.o_Struct, id, in, inRep);
  }

  public static boolean canBeDefIn(CompType type) {
    switch (type.objectKind()) {
      case ObjectsKind.o_Struct:
      case ObjectsKind.o_Enum:
      case ObjectsKind.o_Union:
        return true;
      default: return false;
    }
  }

  public boolean addObject(CompContained o) {
    ((CompContained) o).getIdent().version = id.version;
    return super.addObject(o);
  }

  public void checkConsist(EnumList props, CompRepository rep) throws CDLExceptCheck, CDLExceptLock, CDLExceptRemote {
    CompContained[] cont = contents();
    for (int i=0;i<cont.length;i++) {
      cont[i].checkConsist(props, rep);
    }
  }

  public void addToNormal(Container obj, CDLRepository normRep, java.util.ArrayList delayed) throws CDLExceptToNormal, CDLExceptRemote, CDLExceptLock {
    if (id.isin == IDKind.versiontag) {
      try {
        ((Contained)obj).tag(id.tag);
      } catch (RemoteException e) {
        throw new CDLExceptRemote("Remote exception occured: "+e.getMessage());
      } catch (TIRExceptLock e) {
        throw new CDLExceptLock("Repository is locked.");
      } catch (TIRExceptCreate ecr) {
        throw new CDLExceptToNormal("Can't set tag for "+ fullName());
      }
    }
    cont.toFirst();
    try {
      for (int i=0;i<cont.size(); i++) {
        CompContained akt = (CompContained) cont.aktual();
        switch (akt.objectKind()) {
        // *****************
        case ObjectsKind.o_Struct:
          try {
            StructDef str = ((StructDef) obj).create_struct(akt.getIdent().id);
            if (str == null) 
              throw new CDLExceptToNormal("Can't create struct "+((CompContained)akt).fullName());
            ((CompStruct) akt).addToNormal(str, normRep, delayed);
          } catch (TIRExceptCreate ecr) {
            throw new CDLExceptToNormal("Can't create struct "+((CompContained)akt).fullName());
          }
          break;
        // *****************
        case ObjectsKind.o_StructMember:
          try {
            StructMember mem = ((StructDef) obj).create_member(akt.getIdent().name, ((CompStructMember)akt).tp.toNormal(normRep));
            if (mem == null) 
              throw new CDLExceptToNormal("Can't create struct member "+((CompContained)akt).fullName());
          } catch (TIRExceptCreate ecr) {
            throw new CDLExceptToNormal("Can't create struct member "+((CompContained)akt).fullName()+"\n"+ecr.getMessage());
          }
          break;
        // *****************
        case ObjectsKind.o_Enum:
          try {
            EnumDef en = ((StructDef) obj).create_enum(akt.getIdent().id);
            if (en == null) 
              throw new CDLExceptToNormal("Can't create enum "+((CompContained)akt).fullName());
            ((CompEnum) akt).addToNormal(en);
          } catch (TIRExceptCreate ecr) {
            throw new CDLExceptToNormal("Can't create enum "+((CompContained)akt).fullName());
          }
          break;
        // *****************
        case ObjectsKind.o_Union:
          try {
            UnionDef un = ((StructDef) obj).create_union(akt.getIdent().id, ((CompUnion) akt).switch_type.toNormal(normRep));
            if (un == null) 
              throw new CDLExceptToNormal("Can't create union "+((CompContained)akt).fullName());
            ((CompUnion) akt).addToNormal(un, normRep, delayed);
          } catch (TIRExceptCreate ecr) {
            throw new CDLExceptToNormal("Can't create union "+((CompContained)akt).fullName());
          }
          break;

        default:
          throw new CDLExceptToNormal("Unexcepted kind of object"); 
        }
        cont.toNext();
      }
    } catch (RemoteException e) {
      throw new CDLExceptRemote("Remote exception occured: "+e.getMessage());
    } catch (TIRExceptLock e) {
      throw new CDLExceptLock("Repository is locked.");
    }
  }

  /** from CompType */
  public CDLType toNormal(CDLRepository newRep) throws CDLExceptToNormal {
    // never call this
    throw new CDLExceptToNormal("It should never throw");
  }
}
