/* $Id: JavaGenerator.java,v 2.18 2003/01/15 22:38:24 hnetynka Exp $ */
package SOFA.SOFAnode.Made.CodeGen;

import SOFA.SOFAnode.Made.TIR.*;
import java.io.IOException;
import java.rmi.RemoteException;

import javax.xml.parsers.*;
//import org.xml.sax.*;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

/** Class containing methods for generating all "things" in repository to java.
  *
  * @author Petr Hnetynka
  */
public class JavaGenerator {

  ObjectList    allObjects;     // objects to generate
  GenObjectList genObjects;     // generated objects
  CGFileAccess  fs;             // file manipulation
  GenObjectList pregenIfaces;   // pregenerated interface names
  ArchitectureDef[] archToGen;  // architectures to generate
  /** Info about generated files. */
  GenFileList genFiles; 

  boolean typesonly;

  /** Constructor */
  public JavaGenerator(ObjectList allObjects, ArchitectureDef[] archToGen) throws IOException {
    this.allObjects = allObjects;
    this.genObjects = new GenObjectList();
    this.genFiles = new GenFileList();
    this.archToGen = archToGen;

    PrimitiveGeneratedObject.createPrimGenerObjects(genObjects);

    String curDir = System.getProperty("codegen.dir", null);
    boolean rewrite = Boolean.valueOf(System.getProperty("codegen.rewrite", "true")).booleanValue();
    
    if (curDir == null) {  // use current dir
      fs = new CGFileAccess(".", rewrite);
    } else { // use dir specified in the property
      fs = new CGFileAccess(curDir, rewrite);
    }
    typesonly = false;
  }

  /** Constructor 
    * @param typesonly    typesonly property
    */
  public JavaGenerator(ObjectList allObjects, boolean typesonly) throws IOException {
    this.allObjects = allObjects;
    this.genObjects = new GenObjectList();
    this.typesonly = typesonly;

    PrimitiveGeneratedObject.createPrimGenerObjects(genObjects);

    String curDir = System.getProperty("codegen.dir", null);
    boolean rewrite = Boolean.valueOf(System.getProperty("codegen.rewrite", "true")).booleanValue();
    
    if (curDir == null) {  // use current dir
      fs = new CGFileAccess(".", rewrite);
    } else { // use dir specified in the property
      fs = new CGFileAccess(curDir, rewrite);
    }
  }

  /** Run generation of code.
    *
    * @exception GenerationException    generation error (probably error in TIR)
    * @exception IOException  i/o exception
    * @exception RemoteException remote exception
    */
  public void generate() throws RemoteException, GenerException, IOException {
    pregenIfaces = pregenIfaceNames();
    Debug.println("---- begin of pregenIfaces ----");
    Debug.println(pregenIfaces);
    Debug.println("---- end of pregenIfaces ----");
    for (int i=allObjects.size()-1; i>=0 ; i--) {
      Contained current = allObjects.getObject(i);
      CGFileWriter out = null;
      boolean closeOut = true;
      String name = current.get_identification().name();
      if (GeneratedObject.isReservedName(name))
        name = "_"+name;
      AbsoluteName absN = current.get_identification().absolute_name();
      String pkg = CGFileAccess.absToPkg(absN);
      Debug.println("Currently generated object: "+absN.name());
 
      switch (current.get_def_kind().value()) {
        case DefinitionKind.dk_Constant:
          out = fs.newCGWriter(CGFileAccess.absToPath(absN, name,  true));
          generateConstant((ConstantDef) current, out, name, pkg);
          break;
        case DefinitionKind.dk_Enum:
          out = fs.newCGWriter(CGFileAccess.absToPath(absN, name,  true));
          generateEnum((EnumDef) current, out, name, CGFileAccess.absToPath(absN, name, false), pkg);
          break;
        case DefinitionKind.dk_Struct:
          out = fs.newCGWriter(CGFileAccess.absToPath(absN, name,  true));
          generateStruct((StructDef) current, out, name, CGFileAccess.absToPath(absN, name, false), pkg);
          break;
        case DefinitionKind.dk_Exception:
          out = fs.newCGWriter(CGFileAccess.absToPath(absN, name,  true));
          generateException((ExceptionDef) current, out, name, CGFileAccess.absToPath(absN, name, false), pkg);
          break;
        case DefinitionKind.dk_Typedef:
          //out = fs.newCGWriter(CGFileAccess.absToPath(absN, name,  true));
          closeOut = false;
          generateTypedef((TypedefDef) current, name, CGFileAccess.absToPath(absN, name, false), pkg);
          break;
        case DefinitionKind.dk_Interface:
          out = fs.newCGWriter(CGFileAccess.absToPath(absN, name,  true));
          generateInterface((InterfaceDef) current, out, name, CGFileAccess.absToPath(absN, name, false), pkg);
          break;
        default:
          out = fs.newCGWriter(CGFileAccess.absToPath(absN, name, true));
          out.iwriteln("Currently, I don't know, how to generate this object.");
      }
      if (closeOut)
        out.close();
    }
    
    if (!typesonly) {
      // generate component builders
      for (int i=0;i<archToGen.length; i++) {
        CGFileWriter out = null;
        String name = archToGen[i].get_identification().name();
        if (GeneratedObject.isReservedName(name))
          name = "_"+name;
        AbsoluteName absN = archToGen[i].get_identification().absolute_name();
        String pkg = CGFileAccess.archAbsToPkg(absN);
        String provider = absN.elementAt(0);
        out = fs.newCGWriter(CGFileAccess.archAbsToPath(absN, name+"Builder", true));
        generateBuilder(archToGen[i], out, provider, name, pkg);
        out.close();
        generateComponentInfoFile(archToGen[i], absN, name, pkg+"."+name+"Builder");
      }
    }

  }

  /** Pregenerate names of interfaces (neccessary for forward declaration of interfaces).
    */
  private GenObjectList pregenIfaceNames() throws RemoteException {
    JavaMapping jm = new JavaMapping();
    GenObjectList ret = new GenObjectList();
    //System.out.println(allObjects.size());
    for(int i=0; i<allObjects.size(); i++) {
      Contained obj = allObjects.getObject(i);
      if (obj instanceof CDLType) {
        ret.add(new ContainedGeneratedObject(obj, jm.getImplTypeName((CDLType) obj)));
      }
    }
    return ret;
  }

  /** Returns the object with specified key. It searchs in the generated objects and also in the pregenarated objest.
    *
    * @param key    key of the searched object
    * @return the searched object; returns <B>null</B> if the object is not found. 
    */
  private GeneratedObject searchObject(String key) {
    GeneratedObject ret = null;
    if ((ret = genObjects.searchObject(key)) == null) {
      ret = pregenIfaces.searchObject(key);
    }
    return ret;
  }

  /** Return string with concrete number.
    */
  private String getNumber(ExprOperDef expr) throws GenerException, RemoteException, PropException {
    GeneratedObject genObj;
    String key;
    switch (expr.get_eok_kind().value()) {
    case ExprOperKind.eok_boolean: 
      if ( ((ExprBoolean) expr).value() )
        return ("true");
      else
        return ("false");
    case ExprOperKind.eok_char:
      return ( ((ExprChar) expr).toString() );
    case ExprOperKind.eok_constant:
      ExprConstant exprc = (ExprConstant) expr;
      key = GeneratedObject.createKey(exprc.ref_const());
      genObj = searchObject(key);
      if (genObj == null)
        throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+" "+Res.res.getString("MSG_CantBeGenerated"));
      return genObj.genReference(GeneratedObject.INREF);
    case ExprOperKind.eok_double:
      return (Double.toString( ((ExprDouble) expr).value() ));
    case ExprOperKind.eok_fixed:
      ExprFixed exprf = (ExprFixed) expr;
      return ("new java.math.BigDecimal("+exprf.value().toString()+")");
    case ExprOperKind.eok_float:
      return (Float.toString( ((ExprFloat) expr).value() ));
    case ExprOperKind.eok_long:
      return (Long.toString( ((ExprLong) expr).value() ));
    case ExprOperKind.eok_longdouble:
      return (Double.toString( ((ExprLongDouble) expr).value() ));
    case ExprOperKind.eok_longlong:
      return (Long.toString( ((ExprLongLong) expr).value() ));
    case ExprOperKind.eok_octet:
      return (Byte.toString( ((ExprOctet) expr).value() ));
    case ExprOperKind.eok_short:
      return (Short.toString( ((ExprShort) expr).value() ));
    case ExprOperKind.eok_string:
      return ( ((ExprString) expr).value() );
    case ExprOperKind.eok_unsiglong:
      return (Long.toString( ((ExprUnsigLong) expr).value() ));
    case ExprOperKind.eok_unsiglonglong:
      return (Long.toString( ((ExprUnsigLongLong) expr).value() ));
    case ExprOperKind.eok_unsigshort:
      return (Short.toString( ((ExprUnsigShort) expr).value() ));
    case ExprOperKind.eok_wchar:
      return ( ((ExprWchar) expr).toString() );
    case ExprOperKind.eok_wstring:
      return ( ((ExprWstring) expr).value() );
    case ExprOperKind.eok_property:
      throw new PropException(Res.res.getString("MSG_PropInExpr"));
    }
    return "";
  }

  /** Directly write concrete number to the out writer.
    */
  private void writeNumber(ExprOperDef expr, CGFileWriter out) throws GenerException, IOException, RemoteException, PropException {
    GeneratedObject genObj;
    String key;
    switch (expr.get_eok_kind().value()) {
    case ExprOperKind.eok_boolean: 
      if ( ((ExprBoolean) expr).value() )
        out.iwrite("true");
      else
        out.iwrite("false");
      break;
    case ExprOperKind.eok_char:
      out.iwrite( ((ExprChar) expr).value() );
      break;
    case ExprOperKind.eok_constant:
      ExprConstant exprc = (ExprConstant) expr;
      key = GeneratedObject.createKey(exprc.ref_const());
      genObj = searchObject(key);
      if (genObj == null)
        throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+" "+Res.res.getString("MSG_CantBeGenerated"));
      genObj.genReference(out, GeneratedObject.INREF);
      break;
    case ExprOperKind.eok_double:
      out.iwrite(Double.toString( ((ExprDouble) expr).value() ));
      break;
    case ExprOperKind.eok_fixed:
      ExprFixed exprf = (ExprFixed) expr;
      out.iwrite("new java.math.BigDecimal("+exprf.value().toString()+")");
      break;
    case ExprOperKind.eok_float:
      out.iwrite(Float.toString( ((ExprFloat) expr).value() ));
      break;
    case ExprOperKind.eok_long:
      out.iwrite(Long.toString( ((ExprLong) expr).value() ));
      break;
    case ExprOperKind.eok_longdouble:
      out.iwrite(Double.toString( ((ExprLongDouble) expr).value() ));
      break;
    case ExprOperKind.eok_longlong:
      out.iwrite(Long.toString( ((ExprLongLong) expr).value() ));
      break;
    case ExprOperKind.eok_octet:
      out.iwrite(Byte.toString( ((ExprOctet) expr).value() ));
      break;
    case ExprOperKind.eok_short:
      out.iwrite(Short.toString( ((ExprShort) expr).value() ));
      break;
    case ExprOperKind.eok_string:
      out.iwrite( ((ExprString) expr).value() );
      break;
    case ExprOperKind.eok_unsiglong:
      out.iwrite(Long.toString( ((ExprUnsigLong) expr).value() ));
      break;
    case ExprOperKind.eok_unsiglonglong:
      out.iwrite(Long.toString( ((ExprUnsigLongLong) expr).value() ));
      break;
    case ExprOperKind.eok_unsigshort:
      out.iwrite(Short.toString( ((ExprUnsigShort) expr).value() ));
      break;
    case ExprOperKind.eok_wchar:
      out.iwrite( ((ExprWchar) expr).value() );
      break;
    case ExprOperKind.eok_wstring:
      out.iwrite( ((ExprWstring) expr).value() );
      break;
    case ExprOperKind.eok_property:
      throw new PropException(Res.res.getString("MSG_PropInExpr"));
    }
  }

  /** Directly write expression (can be with properties) to the out writer.
    */
  private String expressionWithPropertyToString(ExprOperDef expr) throws GenerException, IOException, RemoteException, PropException {
    GeneratedObject genObj;
    String key;
    StringBuffer ret;
    switch (expr.get_eok_kind().value()) {
    case ExprOperKind.eok_boolean: 
      if ( ((ExprBoolean) expr).value() )
        return "true";
      else
        return "false";
    case ExprOperKind.eok_char:
      return "" + ((ExprChar) expr).value() ;
    case ExprOperKind.eok_constant:
      ExprConstant exprc = (ExprConstant) expr;
      key = GeneratedObject.createKey(exprc.ref_const());
      genObj = searchObject(key);
      if (genObj == null)
        throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+" "+Res.res.getString("MSG_CantBeGenerated"));
      return genObj.genReference(GeneratedObject.INREF);
    case ExprOperKind.eok_double:
      return (Double.toString( ((ExprDouble) expr).value() ));
    case ExprOperKind.eok_fixed:
      ExprFixed exprf = (ExprFixed) expr;
      return ("new java.math.BigDecimal("+exprf.value().toString()+")");
    case ExprOperKind.eok_float:
      return (Float.toString( ((ExprFloat) expr).value() ));
    case ExprOperKind.eok_long:
      return (Long.toString( ((ExprLong) expr).value() ));
    case ExprOperKind.eok_longdouble:
      return (Double.toString( ((ExprLongDouble) expr).value() ));
    case ExprOperKind.eok_longlong:
      return (Long.toString( ((ExprLongLong) expr).value() ));
    case ExprOperKind.eok_octet:
      return (Byte.toString( ((ExprOctet) expr).value() ));
    case ExprOperKind.eok_short:
      return (Short.toString( ((ExprShort) expr).value() ));
    case ExprOperKind.eok_string:
      return ( ((ExprString) expr).value() );
    case ExprOperKind.eok_unsiglong:
      return (Long.toString( ((ExprUnsigLong) expr).value() ));
    case ExprOperKind.eok_unsiglonglong:
      return (Long.toString( ((ExprUnsigLongLong) expr).value() ));
    case ExprOperKind.eok_unsigshort:
      return (Short.toString( ((ExprUnsigShort) expr).value() ));
    case ExprOperKind.eok_wchar:
      return "" + ((ExprWchar) expr).value() ;
    case ExprOperKind.eok_wstring:
      return ( ((ExprWstring) expr).value() );
    case ExprOperKind.eok_property:
      return ( "_prop_"+((ExprProperty) expr).ref_prop() );
    case ExprOperKind.eok_unminus:
      ret = new StringBuffer("(-");
      ret.append(expressionWithPropertyToString( ((ExprUnOperationDef) expr).operand() ));
      ret.append(")");
      return ret.toString();
    case ExprOperKind.eok_unplus:
      ret = new StringBuffer("(+");
      ret.append(expressionWithPropertyToString( ((ExprUnOperationDef) expr).operand()));
      return ret.append(")").toString();
    case ExprOperKind.eok_or:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append("||");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_xor:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append("^");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_and:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append("&&");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_shr:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append(">>");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_shl:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append(">>");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_plus:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append("+");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_minus:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append("-");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_mul:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append("*");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_div:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append("/");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    case ExprOperKind.eok_mod:
      ret = new StringBuffer("(");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand1()));
      ret.append("%");
      ret.append(expressionWithPropertyToString( ((ExprBinOperationDef) expr).operand2()));
      return ret.append(")").toString();
    }
    return "";
  }
  
  /** Generates expression to the out writer. 
    *
    * @param expr   generated expresion
    * @param out    genrate to this writer
    * @exception GenerationException    generation error (probably error in TIR)
    * @exception IOException  i/o exception
    * @exception RemoteException remote exception
    */
  public void generateExpression(ExprOperDef expr, CGFileWriter out) throws GenerException, IOException, RemoteException {
    try {
      ExprOperDef res = expr.calculate();
      // expression can be generated as a one number
      try {
        writeNumber(res, out);
      } catch (PropException pe) {
        throw new GenerException(pe.getMessage());
      }
      return;
    } catch (TIRExceptCalculate e) {
      ;    // it cannot be calculate (contains properties or enum members....)
    }
  }
  
  /** Genrates constant mapping to the out writer.
    * 
    * @param obj   generated constant
    * @param out   generate to this writer
    * @param name  short name of constanta
    * @param pkg   package name
    * @exception GenerationException    generation error (probably error in TIR)
    * @exception IOException  i/o exception
    * @exception RemoteException remote exception
    */
  public void generateConstant(ConstantDef obj, CGFileWriter out, String name, String pkg) throws GenerException, IOException, RemoteException {
    Container parent = obj.get_defined_in();
    String key = GeneratedObject.createKey(obj.type());
    GeneratedObject type = searchObject(key);
    if (type == null) {
      throw new GenerException(Res.res.getString("MSG_ReferenceTo")+" "+key+" "+Res.res.getString("MSG_CantBeGenerated"));
    }
     
    if (parent instanceof InterfaceDef) {  // constant within interface
      type.genReference(out, GeneratedObject.INREF);
      out.iwritespace(1);
      out.iwrite(name);
      out.iwrite(" = ");
      generateExpression(obj.value(), out);
      out.iwriteln(" ;"); out.iwriteln();
      genObjects.add(new ContainedGeneratedObject(obj, pkg+"."+name)); //!!! konstanty uvnitr interfacu!!!
    } else { 
      out.iwriteln("package "+pkg+";");
      out.iwriteln();
      printHeadComment(contained2string(obj), out);
      out.iwrite("public interface ");
      out.iwrite(name);
      out.iwriteln(" {");
      out.incIndent();
      out.iwrite("public static ");
      type.genReference(out, GeneratedObject.INREF);
      out.iwrite(" value = ");
      generateExpression(obj.value(), out);
      out.iwriteln(" ;");
      out.decIndent();
      out.iwriteln("};");
      genObjects.add(new ContainedGeneratedObject(obj, pkg+"."+name+".value")); 
      genFiles.addNewObject(pkg+"."+name,contained2string(obj)); 
    }
  }

  /** Genrates enum mapping to the out writer.
    * 
    * @param obj       generated enum
    * @param out       generate to this writer
    * @param absName   absolute name of writer without ".java" extension
    * @param shortName short name
    * @param pkg   package name
    * @exception GenerationException    generation error (probably error in TIR)
    * @exception IOException  i/o exception
    * @exception RemoteException remote exception
    */
  public void generateEnum(EnumDef obj, CGFileWriter out, String shortName, String absName, String pkg) throws GenerException, IOException, RemoteException {
    try {
      EnumMember[] membs = obj.members();
      out.iwriteln("package "+pkg+";");
      out.iwriteln();
      printHeadComment(contained2string(obj),out);
      out.iwriteln("public class "+shortName+" implements "+GeneratedObject.stdPackage+"CDLEntity {");
      out.incIndent();
      out.iwriteln("private        int __value;");
      out.iwriteln("private static int __size = "+Integer.toString(membs.length)+";");
      out.iwriteln("private static "+pkg+"."+shortName+"[] __array = new "+pkg+"."+shortName+"[__size];");
      out.iwriteln();
      for (int i=0;i<membs.length; i++) {
        String mname = membs[i].name();
        out.iwriteln("public static final int _"+mname+" = "+Integer.toString(i)+";");
        out.iwriteln("public static final "+pkg+"."+shortName+" "+mname+" = new "+pkg+"."+shortName+"(_"+mname+");");
      }
      out.iwriteln();
      out.iwriteln("public int value () {");
      out.iwriteln("  return __value;");
      out.iwriteln("}");
      out.iwriteln();
      out.iwriteln("public static "+pkg+"."+shortName+" from_int(int value) {");
      out.iwriteln("  if (value >= 0 && value < __size)");
      out.iwriteln("    return __array[value];");
      out.iwriteln("  else");
      out.iwriteln("    throw new "+GeneratedObject.stdPackage+"BAD_PARAM();");
      out.iwriteln("}");
      out.iwriteln();
      out.iwriteln("protected Barvy (int value) {");
      out.iwriteln("  __value = value;");
      out.iwriteln("  __array[__value] = this;");
      out.iwriteln("}");
      out.decIndent();
      out.iwriteln("}");   // end of generating of the enum

//      if (!typesonly) {
        createHolder( absName+"Holder.java", shortName,pkg);
  
        // create enum helper
        CGFileWriter outHelp = fs.newCGWriter(absName+"Helper.java");
        outHelp.iwriteln("package "+pkg+";");
        outHelp.iwriteln();
        printHeadComment(null, outHelp);
        outHelp.iwriteln("abstract public class "+shortName+"Helper {");
        outHelp.incIndent();
        outHelp.iwriteln("public static "+pkg+"."+shortName+" read("+GeneratedObject.stdPackage+"InputStream istream) throws java.io.IOException {");
        outHelp.iwriteln("  return "+pkg+"."+shortName+".from_int(istream.read_long ());");
        outHelp.iwriteln("}");
        outHelp.iwriteln();
        outHelp.iwriteln("public static void write("+GeneratedObject.stdPackage+"OutputStream ostream, "+pkg+"."+shortName+" value) throws java.io.IOException {");
        outHelp.iwriteln("  ostream.write_long (value.value ());");
        outHelp.iwriteln("}");
        outHelp.iwriteln();
        
        outHelp.decIndent();
        outHelp.iwriteln("}");
        outHelp.close();
//      }

      genObjects.add(new ContainedGeneratedObject(obj, pkg+"."+shortName));
      genFiles.addNewObject(pkg+"."+shortName,contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Holder",contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Helper",contained2string(obj));
    } catch (TIRExceptLock e) {
      throw new GenerException(Res.res.getString("MSG_LockedTIR"));
    }
  }

  private class StructMemberInfo {
    String name;
    GeneratedObject type;
    public StructMemberInfo(String name, GeneratedObject type) {
      this.name = name;
      this.type = type;
    }
  }

  /** Genrates struct mapping to the out writer.
    * 
    * @param obj       generated struct
    * @param out       generate to this writer
    * @param absName   absolute name of writer without ".java" extension
    * @param shortName short name
    * @param pkg   package name
    * @exception GenerationException    generation error (probably error in TIR)
    * @exception IOException  i/o exception
    * @exception RemoteException remote exception
    */
  public void generateStruct(StructDef obj, CGFileWriter out, String shortName, String absName, String pkg) throws GenerException, IOException, RemoteException {
    try {
      Contained[] cont = obj.contents(null);
      out.iwriteln("package "+pkg+";");
      out.iwriteln();
      printHeadComment(contained2string(obj),out);
      out.iwriteln("public final class "+shortName+" implements "+GeneratedObject.stdPackage+"CDLEntity {");
      out.incIndent();
      
      String key;
      GeneratedObject type;
      StructMemberInfo smi;
      // array for already generated struct members
      java.util.ArrayList genContent = new java.util.ArrayList(cont.length);
      CGFileWriter outInner;
      Identification idInner;
      String nameInner;
      for (int i=0;i<cont.length;i++) {  // go through all objects defined in the struct
        switch (cont[i].get_def_kind().value()) {
        case DefinitionKind.dk_StructMember:
          StructMember sm = (StructMember) cont[i];
          key = GeneratedObject.createKey(sm.type());
          type = searchObject(key);
          if (type == null) {
            throw new GenerException(Res.res.getString("MSG_ReferenceTo")+" "+key+" "+Res.res.getString("MSG_CantBeGenerated"));
          }
          smi = new StructMemberInfo(sm.name(), type);
          out.iwriteln("public "+smi.type.genReference(GeneratedObject.INREF)+" "+smi.name+" = "+smi.type.defaultValue()+";");
          genContent.add(smi);
          break;
        case DefinitionKind.dk_Struct:
          StructDef strd = (StructDef) cont[i];
          idInner = strd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateStruct(strd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        case DefinitionKind.dk_Exception:
          ExceptionDef excd = (ExceptionDef) cont[i];
          idInner = excd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateException(excd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        case DefinitionKind.dk_Enum:
          EnumDef enmd = (EnumDef) cont[i];
          idInner = enmd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateEnum(enmd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        }
      }

      out.iwriteln();
      out.iwriteln("public "+shortName+"() {}");
      out.iwriteln();
      out.iwrite("public "+shortName+"(");
      for (int i=0;i<genContent.size();i++) {
        smi = (StructMemberInfo) genContent.get(i);
        out.iwrite(smi.type.genReference(GeneratedObject.INREF)+" _"+smi.name);
        if (i!=genContent.size() - 1)
          out.iwrite(", ");
      }
      out.iwriteln(") {");
      out.incIndent();
      for (int i=0;i<genContent.size();i++) {
        smi = (StructMemberInfo) genContent.get(i);
        out.iwriteln(smi.name+" = _"+smi.name+";");
      }
      out.decIndent();
      out.iwriteln("}");
      
      out.decIndent();
      out.iwriteln("}");   // end of generating of the struct

//      if (!typesonly) {
        createHolder( absName+"Holder.java", shortName,pkg);
       
        // create struct helper
        CGFileWriter outHelp = fs.newCGWriter(absName+"Helper.java");
        outHelp.iwriteln("package "+pkg+";");
        outHelp.iwriteln();
        printHeadComment(null, outHelp);
        outHelp.iwriteln("abstract public class "+shortName+"Helper {");
        outHelp.incIndent();
        outHelp.iwriteln("public static "+pkg+"."+shortName+" read("+GeneratedObject.stdPackage+"InputStream istream) throws java.io.IOException {");
        outHelp.iwriteln("  "+shortName+" value = new "+shortName+"();");
        for (int i=0;i<genContent.size();i++) {
          smi = (StructMemberInfo) genContent.get(i);
          outHelp.iwriteln("  "+smi.type.genReadType("value."+smi.name)+";");
        }
        outHelp.iwriteln("  return value;");
        outHelp.iwriteln("}");
        outHelp.iwriteln();
        outHelp.iwriteln("public static void write("+GeneratedObject.stdPackage+"OutputStream ostream, "+pkg+"."+shortName+" value) throws java.io.IOException {");
        for (int i=0;i<genContent.size();i++) {
          smi = (StructMemberInfo) genContent.get(i);
          outHelp.iwriteln("  "+ smi.type.genWriteType("value."+smi.name) +";");
        }
        outHelp.iwriteln("}");
        outHelp.iwriteln();
        
        outHelp.decIndent();
        outHelp.iwriteln("}");
        outHelp.close();
//      }  

      genObjects.add(new ContainedGeneratedObject(obj, pkg+"."+shortName));
      genFiles.addNewObject(pkg+"."+shortName,contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Holder",contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Helper",contained2string(obj));
    } catch (TIRExceptLock e) {
      throw new GenerException(Res.res.getString("MSG_LockedTIR"));
    }
  } 
  
  /** Genrates exception mapping to the out writer.
    * 
    * @param obj       generated exception
    * @param out       generate to this writer
    * @param absName   absolute name of writer without ".java" extension
    * @param shortName short name
    * @param pkg   package name
    * @exception GenerationException    generation error (probably error in TIR)
    * @exception IOException  i/o exception
    * @exception RemoteException remote exception
    */
  public void generateException(ExceptionDef obj, CGFileWriter out, String shortName, String absName, String pkg) throws GenerException, IOException, RemoteException {
    try {
      Contained[] cont = obj.contents(null);
      out.iwriteln("package "+pkg+";");
      out.iwriteln();
      printHeadComment(contained2string(obj),out);
      out.iwriteln("public final class "+shortName+" extends "+GeneratedObject.stdPackage+"UserException {");
      out.incIndent();
      
      String key;
      GeneratedObject type;
      StructMemberInfo smi;
      // array for already generated struct members
      java.util.ArrayList genContent = new java.util.ArrayList(cont.length);
      CGFileWriter outInner;
      Identification idInner;
      String nameInner;
      for (int i=0;i<cont.length;i++) {  // go through all objects defined in the struct
        switch (cont[i].get_def_kind().value()) {
        case DefinitionKind.dk_StructMember:
          StructMember sm = (StructMember) cont[i];
          key = GeneratedObject.createKey(sm.type());
          type = searchObject(key);
          if (type == null) {
            throw new GenerException(Res.res.getString("MSG_ReferenceTo")+" "+key+" "+Res.res.getString("MSG_CantBeGenerated"));
          }
          smi = new StructMemberInfo(sm.name(), type);
          out.iwriteln("public "+smi.type.genReference(GeneratedObject.INREF)+" "+smi.name+" = "+smi.type.defaultValue()+";");
          genContent.add(smi);
          break;
        case DefinitionKind.dk_Struct:
          StructDef strd = (StructDef) cont[i];
          idInner = strd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateStruct(strd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        case DefinitionKind.dk_Exception:
          ExceptionDef excd = (ExceptionDef) cont[i];
          idInner = excd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateException(excd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        case DefinitionKind.dk_Enum:
          EnumDef enmd = (EnumDef) cont[i];
          idInner = enmd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateEnum(enmd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        }
      }

      out.iwriteln();
      out.iwriteln("public "+shortName+"() {");  // default constructor
      out.iwriteln("  super();");
      out.iwriteln("}");    // end of default constructor
      out.iwriteln();
      if (genContent.size()!=0) {
        out.iwrite("public "+shortName+"(");  // consructor
        for (int i=0;i<genContent.size();i++) {
          smi = (StructMemberInfo) genContent.get(i);
          out.iwrite(smi.type.genReference(GeneratedObject.INREF)+" _"+smi.name);
          if (i!=genContent.size() - 1)
            out.iwrite(", ");
        }
        out.iwriteln(") {");
        out.incIndent();
        out.iwriteln("super();");
        for (int i=0;i<genContent.size();i++) {
          smi = (StructMemberInfo) genContent.get(i);
          out.iwriteln(smi.name+" = _"+smi.name+";");
        }
        out.decIndent();
        out.iwriteln("}");  //end of constructor
        out.iwriteln();
      }
      out.iwrite("public "+shortName+"(String __reason");  // full consructor
      for (int i=0;i<genContent.size();i++) {
        out.iwrite(", ");
        smi = (StructMemberInfo) genContent.get(i);
        out.iwrite(smi.type.genReference(GeneratedObject.INREF)+" _"+smi.name);
      }
      out.iwriteln(") {");
      out.incIndent();
      out.iwriteln("super(__reason);");
      for (int i=0;i<genContent.size();i++) {
        smi = (StructMemberInfo) genContent.get(i);
        out.iwriteln(smi.name+" = _"+smi.name+";");
      }
      out.decIndent();
      out.iwriteln("}\n");  //end of full constructor

      out.iwrite("public "+shortName+"(String __reason, Throwable __cause");  // full consructor with cause
      for (int i=0;i<genContent.size();i++) {
        out.iwrite(", ");
        smi = (StructMemberInfo) genContent.get(i);
        out.iwrite(smi.type.genReference(GeneratedObject.INREF)+" _"+smi.name);
      }
      out.iwriteln(") {");
      out.incIndent();
      out.iwriteln("super(__reason, __cause);");
      for (int i=0;i<genContent.size();i++) {
        smi = (StructMemberInfo) genContent.get(i);
        out.iwriteln(smi.name+" = _"+smi.name+";");
      }
      out.decIndent();
      out.iwriteln("}");  //end of full constructor with cause
      
      out.decIndent();
      
      out.decIndent();
      out.iwriteln("}");   // end of generating of the exception

//      if (!typesonly) {
        createHolder( absName+"Holder.java", shortName,pkg);
       
        // create exception helper
        CGFileWriter outHelp = fs.newCGWriter(absName+"Helper.java");
        outHelp.iwriteln("package "+pkg+";");
        outHelp.iwriteln();
        printHeadComment(null, outHelp);
        outHelp.iwriteln("abstract public class "+shortName+"Helper {");
        outHelp.incIndent();
        outHelp.iwriteln("public static "+pkg+"."+shortName+" read("+GeneratedObject.stdPackage+"InputStream istream) throws java.io.IOException {");
        outHelp.iwriteln("  "+shortName+" value = new "+shortName+"();");
        for (int i=0;i<genContent.size();i++) {
          smi = (StructMemberInfo) genContent.get(i);
          outHelp.iwriteln("  "+smi.type.genReadType("value."+smi.name)+";");
        }
        outHelp.iwriteln("  return value;");
        outHelp.iwriteln("}");
        outHelp.iwriteln();
        outHelp.iwriteln("public static void write("+GeneratedObject.stdPackage+"OutputStream ostream, "+pkg+"."+shortName+" value) throws java.io.IOException {");
        for (int i=0;i<genContent.size();i++) {
          smi = (StructMemberInfo) genContent.get(i);
          outHelp.iwriteln("  "+ smi.type.genWriteType("value."+smi.name) +";");
        }
        outHelp.iwriteln("}");
        outHelp.iwriteln();
        
        outHelp.decIndent();
        outHelp.iwriteln("}");
        outHelp.close();
//      }

      genObjects.add(new ContainedGeneratedObject(obj, pkg+"."+shortName));
      genFiles.addNewObject(pkg+"."+shortName,contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Holder",contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Helper",contained2string(obj));
    } catch (TIRExceptLock e) {
      throw new GenerException(Res.res.getString("MSG_LockedTIR"));
    }
  }


  /** Genrates interface mapping to the out writer.
    * 
    * @param obj       generated interface
    * @param out       generate to this writer
    * @param absName   absolute name of writer without ".java" extension
    * @param shortName short name
    * @param pkg   package name
    * @exception GenerationException    generation error (probably error in TIR)
    * @exception IOException  i/o exception
    * @exception RemoteException remote exception
    */
  public void generateInterface(InterfaceDef obj, CGFileWriter out, String shortName, String absName, String pkg) throws GenerException, IOException, RemoteException {
    try {
      out.iwriteln("package "+pkg+";");
      out.iwriteln();
      printHeadComment(contained2string(obj),out);
      InterfaceDef[] base = obj.base_interfaces();
      // begin of head of interface
      out.iwrite("public interface "+shortName+" extends "+shortName+"Operations, ");
      for (int i=0; i< base.length; i++) {
        GeneratedObject bitem = searchObject(GeneratedObject.createKey(base[i]));
        out.iwrite(bitem.genReference(GeneratedObject.INREF));
        out.iwrite(", ");
      }
      out.iwriteln(GeneratedObject.stdPackage+"CDLEntity {");
      // end of head of interface
      out.iwriteln("}");

      // begin of ...Operation genration  inteface
      Contained[] cont = obj.contents(null);
      CGFileWriter outOp = fs.newCGWriter(absName+"Operations.java");
      outOp.iwriteln("package "+pkg+";");
      outOp.iwriteln();
      outOp.iwriteln();
      printHeadComment(contained2string(obj), outOp);
      outOp.iwrite("public interface "+shortName+"Operations ");
      if (base.length != 0) {
        outOp.iwrite("extends ");
        for(int i=0; i<base.length;i++) {
          GeneratedObject bitem = searchObject(GeneratedObject.createKey(base[i]));
          outOp.iwrite(bitem.genReference(GeneratedObject.INREF));
          outOp.iwrite("Operations");
          if (i != base.length-1) {
            outOp.iwrite(", ");
          }
        }
      }
      outOp.iwriteln(" {");
      outOp.incIndent();
      CGFileWriter outInner;
      Identification idInner;
      String nameInner;
      GeneratedObject genObj;
      String key;
      for(int i=0; i< cont.length; i++) {
        switch (cont[i].get_def_kind().value()) {
        case DefinitionKind.dk_Struct:
          StructDef strd = (StructDef) cont[i];
          idInner = strd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateStruct(strd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
       case DefinitionKind.dk_Exception:
          ExceptionDef excd = (ExceptionDef) cont[i];
          idInner = excd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateException(excd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        case DefinitionKind.dk_Typedef:
          TypedefDef typd = (TypedefDef) cont[i];
          idInner = typd.get_identification();
          nameInner = idInner.name();
          generateTypedef(typd, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        case DefinitionKind.dk_Constant:
          ConstantDef cnstd = (ConstantDef) cont[i];
          idInner = cnstd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateConstant(cnstd, outInner, nameInner, pkg+"."+shortName+"Package");
          break;
        case DefinitionKind.dk_Enum:
          EnumDef enmd = (EnumDef) cont[i];
          idInner = enmd.get_identification();
          nameInner = idInner.name();
          outInner = fs.newCGWriter(absName+"Package"+java.io.File.separator+nameInner+".java");
          generateEnum(enmd, outInner, nameInner, absName+"Package"+java.io.File.separator+nameInner, pkg+"."+shortName+"Package");
          break;
        case DefinitionKind.dk_Attribute:
          AttributeDef attrd = (AttributeDef) cont[i];
          idInner = attrd.get_identification();
          nameInner = idInner.name();
          if (GeneratedObject.isReservedName(nameInner)) {
            nameInner = "_"+nameInner;
          }
          key = GeneratedObject.createKey(attrd.type());
          genObj = searchObject(key);
          if (genObj == null) {
            throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+" "+Res.res.getString("MSG_CantBeGenerated"));
          }
          String inRef = genObj.genReference(GeneratedObject.INREF);
          outOp.iwriteln(inRef+" "+nameInner+"();");
          if (attrd.mode().value() == AttrMode.ATTR_NORMAL) {
            outOp.iwriteln("void "+nameInner+"("+inRef+" new"+nameInner+");");
          }
          break;
        case DefinitionKind.dk_Operation:
          OperationDef opd = (OperationDef) cont[i];
          idInner = opd.get_identification();
          nameInner = idInner.name();
          if (GeneratedObject.isReservedName(nameInner)) {
            nameInner = "_"+nameInner;
          }
          key = GeneratedObject.createKey(opd.result());
          genObj = searchObject(key);
          if (genObj == null) {
            throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+" "+Res.res.getString("MSG_CantBeGenerated"));
          }
          outOp.iwrite(genObj.genReference(GeneratedObject.INREF)+" "+nameInner+"(");
          ParamDescr[] params = opd.params();
          for (int j=0; j<params.length ;j++ ) {
            String pName = params[j].name();
            if (GeneratedObject.isReservedName(pName)) {
              pName = "_"+pName;
            }
            key = GeneratedObject.createKey(params[j].type());
            genObj = searchObject(key);
            if (genObj == null) {
              throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+"  "+Res.res.getString("MSG_CantBeGenerated"));
            }
            if (params[j].mode().value() != ParamMode.PARAM_IN ) {
              outOp.iwrite(genObj.genReference(GeneratedObject.OUTREF)+" ");
            } else {
              outOp.iwrite(genObj.genReference(GeneratedObject.INREF)+" ");
            }
            outOp.iwrite(pName);
            if (j!=params.length-1) {
              outOp.iwrite(", ");
            }
          }
          outOp.iwrite(") ");
          ExceptionDef[] excs = opd.exceptions();
          if (excs.length != 0) {
            outOp.iwrite("throws ");
            for (int j=0; j<excs.length; j++) {
              key = GeneratedObject.createKey(excs[j]);
              genObj = searchObject(key);
              if (genObj == null) {
                throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+" "+Res.res.getString("MSG_CantBeGenerated"));
              }
              outOp.iwrite(genObj.genReference(GeneratedObject.INREF));
              if (j != excs.length-1 ) {
                outOp.iwrite(", ");
              }
            }
          }
          outOp.iwriteln(";");
          break;
        }
      }
      outOp.decIndent();
      outOp.iwriteln("}");
      outOp.close();

      // end of ...Operation generation

//      if (!typesonly) {
        createHolder( absName+"Holder.java", shortName,pkg);
        
        // create interface helper
        CGFileWriter outHelp = fs.newCGWriter(absName+"Helper.java");
        outHelp.iwriteln("package "+pkg+";");
        outHelp.iwriteln();
        printHeadComment(null, outHelp);
        outHelp.iwriteln("abstract public class "+shortName+"Helper {");
        outHelp.incIndent();
        outHelp.iwriteln("public static "+pkg+"."+shortName+" read("+GeneratedObject.stdPackage+"InputStream istream) throws java.io.IOException {");
        // !!! doplnit telo metody   read
        outHelp.iwriteln("  return null;");
        outHelp.iwriteln("}");
        outHelp.iwriteln();
        outHelp.iwriteln("public static void write("+GeneratedObject.stdPackage+"OutputStream ostream, "+pkg+"."+shortName+" value) throws java.io.IOException {");
        // !!! doplnit telo metody   write
        outHelp.iwriteln("}");
        outHelp.iwriteln();
  
        outHelp.decIndent();
        outHelp.iwriteln("}");
        outHelp.close();
//      }

      genObjects.add(new ContainedGeneratedObject(obj, pkg+"."+shortName));
      genFiles.addNewObject(pkg+"."+shortName,contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Operations",contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Holder",contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Helper",contained2string(obj));
    } catch (TIRExceptLock e) {
      throw new GenerException(Res.res.getString("MSG_LockedTIR"));
    }
  }



  private class TypedefArrayInfo {
    String length;
    public TypedefArrayInfo(String l) {
      length = l;
    }
  }

  /** generates brackets 
    * @param number  number of brackets
    */
  static String genBrackets(int number) {
    StringBuffer sb = new StringBuffer();
    for (int i =0; i<number;i++) {
      sb.append("[]");
    }
    return sb.toString();
  }

  /** generates for cycle. i.e.  "for(index=0; index &grth; max; index++) {" 
    */
  private String generateFor(String index, String max) {
    return "for (int "+index+" = 0; "+index+" < ("+max+"); "+index+"++) {";
  }

  /** generates array filled brackets. i.e.  name[indexPrefix0][indexPrefix1]...
    */
  private String generateArray(String name, String indexPrefix, int number) {
    StringBuffer sb = new StringBuffer(name);
    for (int i=0; i<number; i++) {
      sb.append("[");
      sb.append(indexPrefix);
      sb.append(Integer.toString(i));
      sb.append("]");
    }
    return sb.toString();
  }

  /** Genrates typedef mapping (only helper and holder).
    * 
    * @param obj       generated typedef
    * @param absName   absolute name of writer without ".java" extension
    * @param shortName short name
    * @param pkg   package name
    * @exception GenerationException    generation error (probably error in TIR)
    * @exception IOException  i/o exception
    * @exception RemoteException remote exception
    */
  public void generateTypedef(TypedefDef obj,  String shortName, String absName, String pkg) throws GenerException, IOException, RemoteException {
//    try {
      
      java.util.ArrayList arrInfo = new java.util.ArrayList();
      java.util.ArrayList seqInfo = new java.util.ArrayList();
      boolean ok = false;
      CDLType lastType = obj.original_type();
      
      TypedefArrayInfo tdi = null;
      while (!ok) {
        switch (lastType.get_def_kind().value()) {
          case DefinitionKind.dk_Array:
            String len; 
            try {
              len = getNumber(((ArrayDef) lastType).length());
            } catch (PropException pe) {
              len = null;
            }
            tdi = new TypedefArrayInfo(len);
            arrInfo.add(tdi);
            lastType = ((ArrayDef) lastType).element_type();
            break;
          case DefinitionKind.dk_Sequence:
            try {
              tdi = new TypedefArrayInfo(getNumber(((SequenceDef) lastType).bound()));
            } catch (PropException pe) {
              ; // cannot be here
            }
            seqInfo.add(tdi);
            lastType = ((SequenceDef) lastType).element_type();
            break;
          default:
            ok = true;
            break;
        }
      }

      String origKey = GeneratedObject.createKey(lastType);
      GeneratedObject generObj = searchObject(origKey);
      String inRef = generObj.genReference(GeneratedObject.INREF);

      if (arrInfo.size() == 0 && seqInfo.size()==0) { //there is no array or sequence in this typedef
        switch (lastType.get_def_kind().value()) {
        case DefinitionKind.dk_Primitive:
        case DefinitionKind.dk_String:
        case DefinitionKind.dk_Wstring:
        case DefinitionKind.dk_Typedef:   // generate nothing, use original types
          genObjects.add(new ContainedGeneratedObject(obj, inRef, generObj.genReference(GeneratedObject.OUTREF)));
          return;
        default:  //  use original types, generate Helper only
          genObjects.add(new ContainedGeneratedObject(obj, inRef, generObj.genReference(GeneratedObject.OUTREF)));
//          if (!typesonly) {
            // generate helper
            CGFileWriter outHelp = fs.newCGWriter(absName+"Helper.java");
            outHelp.iwriteln("package "+pkg+";");
            outHelp.iwriteln();
            printHeadComment(null, outHelp);
            outHelp.iwriteln("abstract public class "+shortName+"Helper {");
            outHelp.incIndent();
            outHelp.iwriteln("public static "+inRef+" read("+GeneratedObject.stdPackage+"InputStream istream) {");
            outHelp.iwriteln("  "+inRef+" value = null;");
            outHelp.iwriteln("  "+generObj.genReadType("value")+";");
            outHelp.iwriteln("  return value;");
            outHelp.iwriteln("}");
            outHelp.iwriteln();
            outHelp.iwriteln("public static void write("+GeneratedObject.stdPackage+"OutputStream ostream, "+inRef+" value) {");
            outHelp.iwriteln("  "+ generObj.genWriteType("value") +";");
            outHelp.iwriteln("}");
            outHelp.iwriteln();
      
            outHelp.decIndent();
            outHelp.iwriteln("}");
            outHelp.close();
//          }
          return;
        }
      }
    
      int bracketNum = arrInfo.size() + seqInfo.size();
//      if (!typesonly) {
        createTypedefHolder( absName+"Holder.java", shortName,pkg, inRef+genBrackets(bracketNum));
        // create helper
        CGFileWriter outHelp = fs.newCGWriter(absName+"Helper.java");
        outHelp.iwriteln("package "+pkg+";");
        outHelp.iwriteln();
        printHeadComment(null, outHelp);
        outHelp.iwriteln("abstract public class "+shortName+"Helper {");
        outHelp.incIndent();
        outHelp.iwriteln("public static "+inRef+genBrackets(bracketNum)+" read("+GeneratedObject.stdPackage+"InputStream istream) throws java.io.IOException {");
        outHelp.incIndent();
        //begin of read method body
        outHelp.iwriteln(inRef+" value"+genBrackets(bracketNum)+" = null;");
        for(int i=0; i<arrInfo.size(); i++) {
          if (((TypedefArrayInfo)arrInfo.get(i)).length == null) {
            outHelp.iwriteln("int _len"+Integer.toString(i)+" = istream.read_long();");
            outHelp.iwriteln(generateArray("value","_o",i)+" = new "+inRef+"[_len"+Integer.toString(i)+"]"+genBrackets(bracketNum-i-1)+";");
            outHelp.iwriteln(generateFor("_o"+Integer.toString(i), generateArray("value","_o",i)+".length"));
          } else {
            outHelp.iwriteln(generateArray("value","_o",i)+" = new "+inRef+"["+((TypedefArrayInfo)arrInfo.get(i)).length+"]"+genBrackets(bracketNum-i-1)+";");
            outHelp.iwriteln(generateFor("_o"+Integer.toString(i), ((TypedefArrayInfo)arrInfo.get(i)).length));
          }
          outHelp.incIndent();
        }
        for(int i=arrInfo.size(); i<bracketNum;i++) {
          outHelp.iwriteln("int _len"+Integer.toString(i)+" = istream.read_long();");
          outHelp.iwriteln(generateArray("value","_o",i)+" = new "+inRef+"[_len"+Integer.toString(i)+"]"+genBrackets(bracketNum-i-1)+";");
          outHelp.iwriteln(generateFor("_o"+Integer.toString(i), generateArray("value","_o",i)+".length"));
          outHelp.incIndent();
        }
        outHelp.iwriteln(generObj.genReadType(generateArray("value", "_o", bracketNum))+";");
        for (int i=0; i<bracketNum; i++) {
          outHelp.decIndent();
          outHelp.iwriteln("}");
        }
        outHelp.iwriteln("return value;");
        // end of read method body
        outHelp.decIndent();
        outHelp.iwriteln("}");
        outHelp.iwriteln();
  
        outHelp.iwriteln("public static void write("+GeneratedObject.stdPackage+"OutputStream ostream, "+inRef+genBrackets(bracketNum)+" value) throws java.io.IOException {");
        outHelp.incIndent();
        // begin of write method body
        for(int i=0; i<arrInfo.size(); i++) {
          if (((TypedefArrayInfo)arrInfo.get(i)).length == null) {
            outHelp.iwriteln("ostream.write_long("+generateArray("value","_i",i-arrInfo.size())+".length);");
            outHelp.iwriteln(generateFor("_i"+Integer.toString(i), generateArray("value", "_i",i-arrInfo.size())+".length"));
          } else {
            outHelp.iwriteln(generateFor("_i"+Integer.toString(i), ((TypedefArrayInfo)arrInfo.get(i)).length));
          }
          outHelp.incIndent();
        }
        for(int i=arrInfo.size(); i<bracketNum;i++) {
          outHelp.iwriteln("ostream.write_long("+generateArray("value","_i",i-arrInfo.size())+".length);");
          outHelp.iwriteln(generateFor("_i"+Integer.toString(i), generateArray("value", "_i",i-arrInfo.size())+".length"));
          outHelp.incIndent();
        }
        outHelp.iwriteln(generObj.genWriteType(generateArray("value", "_i", bracketNum))+";");
        for (int i=0; i<bracketNum; i++) {
          outHelp.decIndent();
          outHelp.iwriteln("}");
        }
        // end of write method body
        outHelp.decIndent();
        outHelp.iwriteln("}");
        outHelp.iwriteln();
  
        outHelp.decIndent();
        outHelp.iwriteln("}");
        outHelp.close();
//      }
     
      genObjects.add(new ContainedGeneratedObject(obj, inRef+genBrackets(bracketNum), pkg+"."+shortName+"Holder"));
      genFiles.addNewObject(pkg+"."+shortName+"Holder",contained2string(obj));
      genFiles.addNewObject(pkg+"."+shortName+"Helper",contained2string(obj));
      
 //   } catch (TIRExceptLock e) {
 //     throw new GenerException(Res.res.getString("MSG_LockedTIR"));
 //   }
  }




  

  /** Print head comment (date, etc.) to writer.
    * @param out   output writer
    * @param tirID textual form of identification from TIR
    * @exception IOException  i/o exception
    */
  private void printHeadComment(String tirID, CGFileWriter out) throws IOException {
    java.util.Date date = java.util.Calendar.getInstance().getTime();
    java.text.DateFormat df = java.text.DateFormat.getDateTimeInstance(java.text.DateFormat.FULL, java.text.DateFormat.FULL);
    out.iwriteln("/**");
    out.iwriteln("  * "+Res.res.getString("MSG_GeneratedBy")+" SOFA CodeGen");
    if (tirID != null) {
      out.iwriteln("  * "+tirID);
    }
    out.iwriteln("  * "+df.format(date));
    out.iwriteln("  */");
    out.iwriteln();
  }

  /** Creates holder for given name.
    *
    * @param fileName    name of file
    * @param name        short name without Helper sufix
    * @param pkg         package
    * @exception IOException  i/o exception
    */
  private void createHolder(String fileName, String name, String pkg) throws IOException {
    CGFileWriter out = fs.newCGWriter(fileName);
    out.iwriteln("package "+pkg+";");
    out.iwriteln();
    printHeadComment(null,out);
    out.iwriteln("public class "+name+"Holder implements "+GeneratedObject.stdPackage+"Streamable, java.io.Serializable {");
    out.incIndent();
    out.iwriteln("public "+pkg+"."+name+" value = null;");
    out.iwriteln();
    out.iwriteln("public "+name+"Holder() {}");
    out.iwriteln();
    out.iwriteln("public "+name+"Holder("+pkg+"."+name+" initialValue) {");
    out.iwriteln("  value = initialValue;");
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void _read("+GeneratedObject.stdPackage+"InputStream i) throws java.io.IOException {");
    out.iwriteln("  "+pkg+"."+name+"Helper.read(i);");
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void _write("+GeneratedObject.stdPackage+"OutputStream o) throws java.io.IOException {");
    out.iwriteln("  "+pkg+"."+name+"Helper.write(o, value);");
    out.iwriteln("}");
    out.decIndent();
    out.iwriteln("}");
    out.close();
  }

 /** Creates holder for given name of typedef.
    *
    * @param fileName    name of file
    * @param name        short name without Holder sufix
    * @param pkg         package
    * @param type        read / write type
    * @exception IOException  i/o exception
    */
  private void createTypedefHolder(String fileName, String name, String pkg, String type) throws IOException {
    CGFileWriter out = fs.newCGWriter(fileName);
    out.iwriteln("package "+pkg+";");
    out.iwriteln();
    printHeadComment(null,out);
    out.iwriteln("public class "+name+"Holder implements "+GeneratedObject.stdPackage+"Streamable, java.io.Serializable {");
    out.incIndent();
    out.iwriteln("public "+type+" value = null;");
    out.iwriteln();
    out.iwriteln("public "+name+"Holder() {}");
    out.iwriteln();
    out.iwriteln("public "+name+"Holder("+type+" initialValue) {");
    out.iwriteln("  value = initialValue;");
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void _read("+GeneratedObject.stdPackage+"InputStream i) throws java.io.IOException {");
    out.iwriteln("  "+pkg+"."+name+"Helper.read(i);");
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void _write("+GeneratedObject.stdPackage+"OutputStream o) throws java.io.IOException {");
    out.iwriteln("  "+pkg+"."+name+"Helper.write(o, value);");
    out.iwriteln("}");
    out.decIndent();
    out.iwriteln("}");
    out.close();
  }

  private String contained2string(Contained obj) throws RemoteException {
    Identification id = obj.get_identification();
    StringBuffer sb = new StringBuffer();
    sb.append(id.absolute_name().name());
    sb.append("?");
    sb.append(id.version());
    return sb.toString();
  }

  private void generateBuilder(ArchitectureDef arch, CGFileWriter out, String provider, String name, String pkg) throws IOException, RemoteException, GenerException {
    Contained[] insts = null;
    boolean isPrimitive = false;
    java.util.ArrayList instInfos = null;
    boolean isSystem = arch.is_system();
    try {
      insts = arch.contents(((Repository) arch.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Inst));
      if (insts.length == 0)
        isPrimitive = true;
    } catch (TIRExceptLock e) {
      throw new GenerException(Res.res.getString("MSG_LockedTIR"));
    }
    
  
    out.iwriteln("package "+pkg+";");
    out.iwriteln();
    printHeadComment("Component builder "+Res.res.getString("MSG_For")+" "+contained2string(arch),out);
    out.iwriteln("public class "+name+"Builder extends SOFA.Component.DCUP.ComponentBuilderImpl {");
    out.incIndent();
    out.iwriteln("// "+Res.res.getString("MSG_DefinedInAncestor"));
    out.iwriteln("// protected SOFA.Component.ComponentManager cm;");
    out.iwriteln("// protected SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor dd;");
    out.iwriteln();
    if (isPrimitive) {
      out.iwriteln(name+"Impl impl;");
      out.iwriteln();
    } else {
      instInfos = new java.util.ArrayList();
      // generating: obtaining of dd of subcomponents
      for (int i=0; i<insts.length; i++) {
        InstInfo iInfo = InstInfo.getInstInfo(insts[i]);
        instInfos.add(iInfo);
        out.iwriteln("SOFA.Component.DCUP.DCUPComponentManager "+iInfo.name+"cm"+genBrackets(iInfo.arrays.size())+" = null;");
      }
    }
    generateProperties(arch, out);
    out.iwriteln();
    out.iwriteln("public "+name+"Builder() {");
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void initialize(SOFA.Component.ComponentManager cm, SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor dd) {");
    out.incIndent();
    out.iwriteln("super.initialize(cm, dd);");
    try {
      generateBuilderInitialize(arch, insts, instInfos, out, name ,isPrimitive, isSystem);
    } catch (PropException e) {
      throw new GenerException(e.getMessage());
    }
    out.decIndent();
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void onLeaving(SOFA.SOFAnode.Run.Storage stateStore) throws SOFA.Component.ComponentLifecycleException {");
    out.iwriteln("  if (!stopped) {");
    out.iwriteln("    // "+Res.res.getString("MSG_AddFinCode"));
    out.iwriteln("  } else {");
    out.iwriteln("    SOFA.SOFAnode.Run.Deployment.DeplDock dock = cm.getLocalDeplDock();");
    out.iwriteln("    try {");
    out.iwriteln("      dock.removeComponent(fullInstanceName);");
    out.iwriteln("    } catch (SOFA.SOFAnode.Run.Deployment.DeploymentException e) {");
    out.iwriteln("      throw new SOFA.Component.ComponentLifecycleException(\"Component can't be removed\", e);");
    out.iwriteln("    }");
    out.iwriteln("  }");
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void onArrival(SOFA.SOFAnode.Run.Storage stateStore) throws SOFA.Component.ComponentLifecycleException {");
    try {
      generateBuilderOnArrival(arch, insts, instInfos, out, name ,isPrimitive, isSystem);
    } catch (PropException e) {
      throw new GenerException(e.getMessage());
    }
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void store(SOFA.SOFAnode.Run.Storage stateStore) throws SOFA.Component.ComponentLifecycleException {");
    out.iwriteln("");
    out.iwriteln("}");
    out.iwriteln();
    out.iwriteln("public void restore(SOFA.SOFAnode.Run.Storage stateStore) throws SOFA.Component.ComponentLifecycleException {");
    out.iwriteln("");
    out.iwriteln("}");
    out.decIndent();
    out.iwriteln("}");
  }

  /** Generates definition of the CDL properties and inital setter method. It is called from
    * generateBuilder. 
    * @param arch   architecture
    * @param out    output
    * @exception IOException
    * @exception RemoteException
    * @exception GenerException
    */
  private void generateProperties(ArchitectureDef arch, CGFileWriter out) throws IOException, RemoteException, GenerException {
   try {
      java.util.Hashtable generatedNames = new java.util.Hashtable ();
      Contained[] archProps = arch.contents(((Repository) arch.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Property));
      Contained[] frameProps = arch.frame().contents(((Repository) arch.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Property));
      for (int i=0; i<frameProps.length; i++) {
        PropertyDef prop = (PropertyDef) frameProps[i];
        String name = prop.get_identification().name();
        String key = GeneratedObject.createKey(prop.type());
        GeneratedObject generObj = searchObject(key);
        if (generObj == null) {
          throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+" "+Res.res.getString("MSG_CantBeGenerated"));
        }
        String inRef = generObj.genReference(GeneratedObject.INREF);
        generatedNames.put(name, inRef);
        out.iwriteln("public "+inRef+" _prop_"+name+";");
      }
      out.iwriteln("");
      for (int i=0; i<archProps.length; i++) {
        PropertyDef prop = (PropertyDef) archProps[i];
        String name = prop.get_identification().name();
        if (generatedNames.containsKey(name))
          break;
        String key = GeneratedObject.createKey(prop.type());
        GeneratedObject generObj = searchObject(key);
        if (generObj == null) {
          throw new GenerException(Res.res.getString("MSG_TIRError")+"\n"+key+" "+Res.res.getString("MSG_CantBeGenerated"));
        }
        String inRef = generObj.genReference(GeneratedObject.INREF);
        generatedNames.put(name, inRef);
        out.iwriteln("public "+inRef+" _prop_"+name+";");
      }
      out.iwriteln("");
      out.iwriteln("private void initProperties() throws NumberFormatException {");
      out.incIndent();
      for (java.util.Enumeration e = generatedNames.keys(); e.hasMoreElements(); ) {
        String name = (String) e.nextElement();
        out.iwrite((String) generatedNames.get(name));
        out.iwriteln(" _prop_"+name+" = "+baseTypeToParseMethod((String)generatedNames.get(name), "dd.getProperty(\""+name+"\")")+";");
      }
      out.decIndent();
      out.iwriteln("}\n");
    } catch (TIRExceptLock e) {
      throw new GenerException(Res.res.getString("MSG_LockedTIR"));
    }
  }

  private void generateBuilderInitialize(ArchitectureDef arch, Contained[] insts, java.util.ArrayList instInfos, CGFileWriter out, String name, boolean isPrimitive, boolean isSystem) throws RemoteException, IOException, GenerException, PropException {
    out.iwriteln("initProperties();");
    out.iwriteln("");
    if (isPrimitive) {
      out.iwriteln("impl = new "+name+"Impl();");
      // out.iwriteln("// "+Res.res.getString("MSG_AddCodeProvReq"));
      out.iwriteln("");
      generateBuilderInitializeBinds(arch, insts, instInfos, out, name ,isPrimitive, isSystem);
      out.iwriteln("");
      out.iwriteln("// "+Res.res.getString("MSG_ChangeIfIsUpdatable"));
      out.iwriteln("// updatable = false;");
    } else {
      out.iwriteln("SOFA.SOFAnode.Run.Deployment.DeplDock dock = cm.getLocalDeplDock();");
      // generating: obtaining of dd of subcomponents
      for (int i=0; i<insts.length; i++) {
        InstInfo iInfo = (InstInfo) instInfos.get(i);
        out.iwriteln("SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor "+iInfo.name+"dd = dd.getSubComponentDeploymentDescriptor(\""+iInfo.name+"\");");
      }
      out.iwriteln("");
      
      // generating:  instantiation of subcomponents
      out.iwriteln("String location = null;");
      int bracketNum;
      for(int i=0; i<instInfos.size(); i++) {
        InstInfo iInfo = (InstInfo) instInfos.get(i);
        out.iwriteln("SOFA.Connector.Reference "+iInfo.name+"ref"+genBrackets(iInfo.arrays.size())+" = null;");
//        out.iwriteln("SOFA.Component.DCUP.DCUPComponentManager "+iInfo.name+"cm"+genBrackets(iInfo.arrays.size())+" = null;");
        out.iwriteln("location = dd.getSubComponentLocation(\""+iInfo.name+"\");");
        out.iwriteln("if (SOFA.Util.SOFAURI.isLocal(location)) {");
        out.incIndent();
        out.iwriteln("try {");
        out.incIndent();
        bracketNum = iInfo.arrays.size();
        for (int j=0; j<bracketNum; j++) {
          out.iwriteln(generateArray(iInfo.name+"ref","_o",j)+" = new SOFA.Connector.Reference["+expressionWithPropertyToString((ExprOperDef)iInfo.arrays.get(j))+"]"+genBrackets(bracketNum-j-1)+";");
          out.iwriteln(generateArray(iInfo.name+"cm","_o",j)+" = new SOFA.Component.DCUP.DCUPComponentManager["+expressionWithPropertyToString((ExprOperDef)iInfo.arrays.get(j))+"]"+genBrackets(bracketNum-j-1)+";");
          out.iwriteln(generateFor("_o"+Integer.toString(j), expressionWithPropertyToString((ExprOperDef)iInfo.arrays.get(j))));
          out.incIndent();
        }
        out.iwriteln("String instName = "+generateInstName(iInfo.name,"_o",bracketNum)+";");
        out.iwriteln( generateArray(iInfo.name+"ref","_o",bracketNum)+" = dock.instantiate("+iInfo.name+"dd, instName);");
        out.iwriteln( generateArray(iInfo.name+"cm","_o",bracketNum)+" = getCMConnectorFromRef("+generateArray(iInfo.name+"ref","_o",bracketNum)+");");
        out.iwriteln("cm.registerSubcomponent(instName, "+generateArray(iInfo.name+"ref","_o",bracketNum)+");");
        for (int j=0; j<bracketNum; j++) {
          out.decIndent();
          out.iwriteln("}");
        }
        out.decIndent();
        out.iwriteln("} catch (SOFA.SOFAnode.Run.Deployment.DeploymentException e) {");
        out.iwriteln("  throw new SOFA.Component.ComponentLifecycleException(\"Can't create subcomponent CMs\", e);");
        out.iwriteln("} catch (SOFA.Component.InstantiationException e) {");
        out.iwriteln("  throw new SOFA.Component.ComponentLifecycleException(\"Can't obtain connectors to subcomponent CMs\", e);");
        out.iwriteln("} catch (SOFA.Component.NamingException e) {");
        out.iwriteln("  throw new SOFA.Component.ComponentLifecycleException(\"Can't register subcomponent\", e);");
        out.iwriteln("}");
        out.decIndent();
        out.iwriteln("} else {");
        out.incIndent();
        out.iwriteln("try {");
        out.incIndent();
        out.iwriteln("SOFA.Connector.Reference subDDockRef = SOFA.SOFAnode.Run.Deployment.DeplDockImpl.bytesToSofaReference(SOFA.Util.SOFAURI.getReferenceToDeplDockFromSOFAURI(location));");
        out.iwriteln("SOFA.SOFAnode.Run.Deployment.DeplDock subDDock = (SOFA.SOFAnode.Run.Deployment.DeplDock) SOFA.Connector.Boot.DeplDockConnector.createClt(subDDockRef);");
        bracketNum = iInfo.arrays.size();
        for (int j=0; j<bracketNum; j++) {
          out.iwriteln(generateArray(iInfo.name+"ref","_o",j)+" = new SOFA.Connector.Reference["+expressionWithPropertyToString((ExprOperDef)iInfo.arrays.get(j))+"]"+genBrackets(bracketNum-j-1)+";");
          out.iwriteln(generateArray(iInfo.name+"cm","_o",j)+" = new SOFA.Component.DCUP.DCUPComponentManager["+expressionWithPropertyToString((ExprOperDef)iInfo.arrays.get(j))+"]"+genBrackets(bracketNum-j-1)+";");
          out.iwriteln(generateFor("_o"+Integer.toString(j), expressionWithPropertyToString((ExprOperDef)iInfo.arrays.get(j))));
          out.incIndent();
        }
        out.iwriteln("String instName = "+generateInstName(iInfo.name,"_o",bracketNum)+";");
        out.iwriteln( generateArray(iInfo.name+"ref","_o",bracketNum)+" = subDDock.instantiate("+iInfo.name+"dd, instName);");
        out.iwriteln( generateArray(iInfo.name+"cm","_o",bracketNum)+" = getCMConnectorFromRef("+generateArray(iInfo.name+"ref","_o",bracketNum)+");");
        out.iwriteln("cm.registerSubcomponent(instName, "+generateArray(iInfo.name+"ref","_o",bracketNum)+");");
        for (int j=0; j<bracketNum; j++) {
          out.decIndent();
          out.iwriteln("}");
        }
        out.decIndent();
        out.iwriteln("} catch (SOFA.Connector.ConnectorException e) {");
        out.iwriteln("  throw new SOFA.Component.ComponentLifecycleException(\"Can't create subcomponent CMs\", e);");
        out.iwriteln("} catch (SOFA.Util.SOFAURI.BadSOFAURIException e) {");
        out.iwriteln("  throw new SOFA.Component.ComponentLifecycleException(\"Can't create subcomponent CMs\", e);");
        out.iwriteln("} catch (SOFA.SOFAnode.Run.Deployment.DeploymentException e) {");
        out.iwriteln("  throw new SOFA.Component.ComponentLifecycleException(\"Can't create subcomponent CMs\", e);");
        out.iwriteln("} catch (SOFA.Component.InstantiationException e) {");
        out.iwriteln("  throw new SOFA.Component.ComponentLifecycleException(\"Can't obtain connectors to subcomponent CMs\", e);");
        out.iwriteln("} catch (SOFA.Component.NamingException e) {");
        out.iwriteln("  throw new SOFA.Component.ComponentLifecycleException(\"Can't register subcomponent\", e);");
        out.iwriteln("}");
        out.decIndent();
        out.iwriteln("}");
        out.iwriteln("");
      }
      out.iwriteln("");
      // out.iwriteln("// "+Res.res.getString("MSG_AddCodeForBinds"));
      out.iwriteln("");
      generateBuilderInitializeBinds(arch, insts, instInfos, out, name ,isPrimitive, isSystem);
      out.iwriteln("");
      out.iwriteln("// "+Res.res.getString("MSG_ChangeIfIsUpdatable"));
      out.iwriteln("// updatable = false;");
    }
  }

  /** Generates onArrival method of the builder.
    */
  private void generateBuilderOnArrival(ArchitectureDef arch, Contained[] insts, java.util.ArrayList instInfos, CGFileWriter out, String name, boolean isPrimitive, boolean isSystem) throws RemoteException, IOException, GenerException, PropException {
    if (isPrimitive) {
      out.iwriteln("");
    } else {
      int bracketNum;
      out.incIndent();
      for(int i=0; i<instInfos.size(); i++) {
        InstInfo iInfo = (InstInfo) instInfos.get(i);
        bracketNum = iInfo.arrays.size();
        if (bracketNum == 0) { // inst is not array
          out.iwriteln(iInfo.name+"cm.createComponent(stateStore);");
        } else { // inst is array
          for (int j=0; j<bracketNum; j++) {
            out.iwriteln(generateFor("_i"+Integer.toString(j), expressionWithPropertyToString((ExprOperDef)iInfo.arrays.get(j))));
            out.incIndent();
          }
          out.iwriteln(generateArray(iInfo.name+"cm","_i",bracketNum)+".createComponent(stateStore);");
          for (int j=0; j<bracketNum; j++) {
            out.decIndent();
            out.iwriteln("}");
          }
        }
      }
      out.decIndent();
    }
  }

  /** Generates code for bindings in the builder initialize method.
    */
  private void generateBuilderInitializeBinds(ArchitectureDef arch, Contained[] insts, java.util.ArrayList instInfos, CGFileWriter out, String name, boolean isPrimitive, boolean isSystem) throws RemoteException, IOException, GenerException, PropException {
  
    Contained[] binds;
    java.util.ArrayList delegs = new java.util.ArrayList();
    java.util.ArrayList subsums = new java.util.ArrayList();
    java.util.ArrayList bindspr = new java.util.ArrayList();
    try {
      binds = arch.contents(((Repository) arch.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Bind));
    } catch (TIRExceptLock e) {
      throw new GenerException(Res.res.getString("MSG_LockedTIR"));
    }

    // bind sub1:req to sub2:prov
    if (binds != null && binds.length != 0) {
      out.iwriteln("try {");
      out.incIndent();
      out.iwriteln("SOFA.Connector.Reference pref;");
      for (int i=0; i<binds.length; i++) {
        BindDef bind = (BindDef) binds[i];
        switch (bind.mode().value()) {
          case BindMode.DELEGATE:
            delegs.add(bind);
            break;
          case BindMode.SUBSUME:
            subsums.add(bind);
            break;
          case BindMode.BIND:
            if (bind.lhsop().get_bo_kind().value() == BindOperKind.bok_provreq) {
              bindspr.add(bind);
            } else {
              BindOperSub lhs = (BindOperSub) bind.lhsop();
              BindOperSub rhs = (BindOperSub) bind.rhsop();
              BindType[] lelems = lhs.elements();
              BindType[] lsubs = lhs.subs();
              BindType[] relems = rhs.elements();
              BindType[] rsubs = rhs.subs();
              String lcompname;
              String rcompname;
              boolean containsArray = false;
              if (lsubs[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                lcompname = ((BindTypeNormal) lsubs[0]).name();
              } else {
                lcompname = ((BindTypeArray) lsubs[0]).name();
                containsArray = true;
              }
              if (rsubs[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                rcompname = ((BindTypeNormal) rsubs[0]).name();
              } else {
                rcompname = ((BindTypeArray) rsubs[0]).name();
                containsArray = true;
              }
              String lelemname;
              String relemname;
              if (lelems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                lelemname = ((BindTypeNormal) lelems[0]).name();
              } else {
                lelemname = ((BindTypeArray) lelems[0]).name();
                containsArray = true;
              }
              if (relems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                relemname = ((BindTypeNormal) relems[0]).name();
              } else {
                relemname = ((BindTypeArray) relems[0]).name();
                containsArray = true;
              }
              if (containsArray) {
                out.iwriteln(Res.res.getString("MSG_ProvideArray"));
              }
              out.iwriteln("pref = "+rcompname+"cm.getProvisionReference(\""+relemname+"\");");
              out.iwriteln(lcompname+"cm.setRequirement(\""+lelemname+"\", pref);");
            }
            break;
        }
      }
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.NamingException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantBindSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("}");
    }

    // delegate prov to sub:prov
    if (delegs.size() != 0) {
      out.iwriteln("try {");
      out.incIndent();
      out.iwriteln("SOFA.Connector.Reference pref;");
      for (int i=0; i<delegs.size(); i++) {
        BindDef bind = (BindDef) delegs.get(i);
        BindOperProvReq lhs = (BindOperProvReq) bind.lhsop();
        BindOperSub rhs = (BindOperSub) bind.rhsop();
        BindType[] lelems = lhs.elements();
        BindType[] relems = rhs.elements();
        BindType[] rsubs = rhs.subs();
        String rcompname;
        boolean containsArray = false;
        if (rsubs[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
          rcompname = ((BindTypeNormal) rsubs[0]).name();
        } else {
          rcompname = ((BindTypeArray) rsubs[0]).name();
          containsArray = true;
        }
        String lelemname;
        String relemname;
        if (lelems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
          lelemname = ((BindTypeNormal) lelems[0]).name();
        } else {
          lelemname = ((BindTypeArray) lelems[0]).name();
          containsArray = true;
        }
        if (relems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
          relemname = ((BindTypeNormal) relems[0]).name();
        } else {
          relemname = ((BindTypeArray) relems[0]).name();
          containsArray = true;
        }
        if (containsArray) {
          out.iwriteln(Res.res.getString("MSG_ProvideArray"));
        }
        out.iwriteln("pref = "+rcompname+"cm.getProvisionReference(\""+relemname+"\");");
        out.iwriteln("((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).delegateProvision(\""+lelemname+"\", pref);");
      }
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.NamingException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantDelegateSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.InstantiationException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantDelegateSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("}");
    }

    // subsume sub:req to req
    if (subsums.size() != 0) {
           // instances on which the "connectSubsumable" must be called
      java.util.Hashtable subsInsts = new java.util.Hashtable();
      out.iwriteln("try {");
      out.incIndent();
      out.iwriteln("SOFA.Connector.Reference pref;");
      for (int i=0; i<subsums.size(); i++) {
        BindDef bind = (BindDef) subsums.get(i);
        BindOperSub lhs = (BindOperSub) bind.lhsop();
        BindOperProvReq rhs = (BindOperProvReq) bind.rhsop();
        BindType[] lelems = lhs.elements();
        BindType[] lsubs = lhs.subs();
        BindType[] relems = rhs.elements();
        String lcompname;
        boolean containsArray = false;
        if (lsubs[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
          lcompname = ((BindTypeNormal) lsubs[0]).name();
        } else {
          lcompname = ((BindTypeArray) lsubs[0]).name();
          containsArray = true;
        }
        String lelemname;
        String relemname;
        if (lelems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
          lelemname = ((BindTypeNormal) lelems[0]).name();
        } else {
          lelemname = ((BindTypeArray) lelems[0]).name();
          containsArray = true;
        }
        if (relems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
          relemname = ((BindTypeNormal) relems[0]).name();
        } else {
          relemname = ((BindTypeArray) relems[0]).name();
          containsArray = true;
        }
        if (containsArray) {
          out.iwriteln(Res.res.getString("MSG_ProvideArray"));
        }
        out.iwriteln("pref = ((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).createSubsumableRequirement(\""+relemname+"\");");
        out.iwriteln(lcompname+"cm.setRequirement(\""+lelemname+"\", pref);");
        subsInsts.put("lcompname", "");
      }
      out.iwriteln("");
      for (java.util.Enumeration e = subsInsts.keys() ; e.hasMoreElements() ;) {
        String instName = (String) e.nextElement();
        out.iwriteln(instName+"cm.connectSubsumable();");
      }
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.NamingException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantSubsumeSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.InstantiationException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantSubsumeSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("}");
    }

    // bind prov to req
    if (bindspr.size() != 0) {
      out.iwriteln("try {");
      out.incIndent();
      out.iwriteln("SOFA.Connector.Reference pref;");
      for (int i=0; i<bindspr.size(); i++) {
        BindDef bind = (BindDef) bindspr.get(i);
        boolean containsArray = false;
        BindOperProvReq lhs = (BindOperProvReq) bind.lhsop();
        BindOperProvReq rhs = (BindOperProvReq) bind.rhsop();
        BindType[] lelems = lhs.elements();
        BindType[] relems = rhs.elements();
        String lelemname;
        String relemname;
        if (lelems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
          lelemname = ((BindTypeNormal) lelems[0]).name();
        } else {
          lelemname = ((BindTypeArray) lelems[0]).name();
          containsArray = true;
        }
        if (relems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
          relemname = ((BindTypeNormal) relems[0]).name();
        } else {
          relemname = ((BindTypeArray) relems[0]).name();
          containsArray = true;
        }
        if (containsArray) {
          out.iwriteln(Res.res.getString("MSG_ProvideArray"));
        }
        out.iwriteln("pref = ((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).createSubsumableRequirement(\""+relemname+"\");");
        out.iwriteln("((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).delegateProvision(\""+lelemname+"\", pref);");
      }
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.NamingException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantDirectBindSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.InstantiationException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantDirectbindSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("}");
    }
    
    // provisions 
    Contained[] provs = null;
    try {
      provs = arch.frame().contents(((Repository) arch.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Provides));
    } catch (TIRExceptLock e) {
      throw new GenerException(Res.res.getString("MSG_LockedTIR"));
    }
    for(int i=0; i<provs.length; i++) {
      String prName = provs[i].get_identification().name();
      out.iwriteln("// "+Res.res.getString("MSG_BeginOfProvision")+" "+prName);
      if (isPrimitive) {
        out.iwriteln("try {");
        out.incIndent();
        out.iwriteln("((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).reSetProvisionImpl(\""+prName+"\", impl);");
        out.decIndent();
        out.iwriteln("} catch (SOFA.Component.NamingException e) {");
        out.incIndent();
        out.iwriteln("try {");
        out.incIndent();
        out.iwriteln("((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).setProvisionImpl(\""+prName+"\", impl);");
        out.decIndent();
        out.iwriteln("} catch (SOFA.Component.NamingException e1) {");
        out.incIndent();
        out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantRegisterProvision")+"\", e1);");
        out.decIndent();
        out.iwriteln("} catch (SOFA.Component.InstantiationException e1) {");
        out.incIndent();
        out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantCreateProvision")+"\", e1);");
        out.decIndent();
        out.iwriteln("}");
        out.decIndent();
        out.iwriteln("}");
      } else {  // it is not primitive component
        out.iwriteln("try {");
        out.incIndent();
        out.iwriteln("((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).createProvision(\""+prName+"\");");
        out.decIndent();
        out.iwriteln("} catch (SOFA.Component.NamingException e) {");
        out.incIndent();
        out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantRegisterProvision")+"\", e);");
        out.decIndent();
        out.iwriteln("} catch (SOFA.Component.InstantiationException e) {");
        out.incIndent();
        out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantCreateProvision")+"\", e);");
        out.decIndent();
        out.iwriteln("}");
      }
      out.iwriteln("// "+Res.res.getString("MSG_EndOfProvision")+" "+prName);
    }

    if ((delegs.size() + bindspr.size()) != 0) {
      out.iwriteln("try {");
      out.incIndent();
      out.iwriteln("((SOFA.Component.DCUP.DCUPComponentManagerImpl) cm).connectDelegated();");
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.NamingException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantDelegateSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("} catch (SOFA.Component.InstantiationException e) {");
      out.incIndent();
      out.iwriteln("throw new SOFA.Component.ComponentLifecycleException(\""+Res.res.getString("MSG_CantDelegateSubcomponents")+"\", e);");
      out.decIndent();
      out.iwriteln("}");
    }
    

  }

  static class InstInfo {
    String name;
    FrameDef frame;
    java.util.ArrayList arrays;
    public InstInfo(String n, FrameDef f, java.util.ArrayList a) {
      name = n;
      frame = f;
      arrays = a;
    }

    public static InstInfo getInstInfo(Contained inst) throws RemoteException {
      int num = 0;
      java.util.ArrayList arrays = new java.util.ArrayList();
      CDLType obj = ((InstDef) inst).type();
      int defKind = obj.get_def_kind().value();
      while (defKind == DefinitionKind.dk_Typedef || defKind == DefinitionKind.dk_Array) {
        if (defKind == DefinitionKind.dk_Typedef) 
          obj = ((TypedefDef) obj).original_type();
        else {
          arrays.add(((ArrayDef) obj).length());
          obj = ((ArrayDef) obj).element_type();
        }
        defKind = obj.get_def_kind().value();
      }
      FrameDef realType = (FrameDef) obj;
      return new InstInfo(inst.get_identification().name(), realType, arrays);
    }

    public static InstInfo findInstInfo(java.util.ArrayList infos, String name) {
      for (int i=0; i<infos.size(); i++) {
        InstInfo f = (InstInfo) infos.get(i);
        if ( f.name.compareTo(name)==0)
          return f;
      }
      return null;
    }
  }
  
  private static String baseTypeToParseMethod(String type, String from) throws GenerException {
    if (type.compareTo("int")==0)
      return "Integer.parseInt("+from+")";
    if (type.compareTo("long")==0)
      return "Long.parseLong("+from+")";
    if (type.compareTo("byte")==0)
      return "Byte.parseByte("+from+")";
    if (type.compareTo("double")==0)
      return "Double.parseDouble("+from+")";
    if (type.compareTo("boolean")==0)
      return "Boolean.valueOf("+from+")";
    if (type.compareTo("short")==0)
      return "Short.parseShort("+from+")";
    if (type.compareTo("String")==0)
      return from;
    throw new GenerException(Res.res.getString("MSG_UnsupportedPropertyType")+type);
  }

  /** Generate code that creates instantion name */
  private static String generateInstName(String name, String indexPrefix, int number) {
    StringBuffer sb = new StringBuffer("\""+name+"\"");
    for (int i=0; i<number; i++) {
      sb.append("+\"_\"+");
      sb.append("Integer.toString("+indexPrefix+Integer.toString(i)+")");
    }
    return sb.toString();
  }

  private void generateComponentInfoFile(ArchitectureDef obj, SOFA.SOFAnode.Made.TIR.AbsoluteName absN, String name, String builderName) throws RemoteException, GenerException, IOException {
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    Document document = null;
    try {
      DocumentBuilder builder = factory.newDocumentBuilder();
      document = builder.newDocument();
      Element root = null;
      if (obj.is_system()) 
        root = (Element) document.createElement("sofa_system"); 
      else
        root = (Element) document.createElement("sofa_component");
      document.appendChild(root);
      Element el;
      Text tx;
      
      el = document.createElement("architecture_ref");
      tx = document.createTextNode(contained2string(obj));
      el.appendChild(tx);
      root.appendChild(el);

      el = document.createElement("frame_ref");
      tx = document.createTextNode(contained2string(obj.frame()));
      el.appendChild(tx);
      root.appendChild(el);

      el = document.createElement("version");
      tx = document.createTextNode("!set.version.here!");
      el.appendChild(tx);
      root.appendChild(el);
      
      Element entit = (Element) document.createElement("cdl_entities");
      for (int i=0; i<genFiles.size(); i++) {
        GenFile go = genFiles.getObject(i);
        el = document.createElement("entity");
        el.setAttribute("cdlname", go.cdlName);
        el.setAttribute("javaname", go.genName);
        entit.appendChild(el);
      }
      root.appendChild(entit);
      
      el = document.createElement("implementation");
      el.setAttribute("type", "builder");
      tx = document.createTextNode(builderName);
      el.appendChild(tx);
      root.appendChild(el);

      try {
        Contained[] cont = obj.contents(((Repository)obj.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Property));
        if (cont!=null && cont.length!=0) {
          for (int i=0; i<cont.length; i++) {
            el = document.createElement("property");
            el.setAttribute("name",cont[i].get_identification().name());
            el.setAttribute("value","!set.property.value.here!");
            root.appendChild(el);
          }
        }
        
        cont = obj.contents(((Repository)obj.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Inst));
        boolean isPrimitive = true;
        if (cont!=null && cont.length!=0) {
          isPrimitive = false;
          for (int i=0; i<cont.length; i++) {
            el = document.createElement("component_ref");
            el.setAttribute("inst",cont[i].get_identification().name());
            el.setAttribute("arch","!set.architecture.name.here!");
            el.setAttribute("version","!set.implementation.version.here!");
            root.appendChild(el);
          }
        }

        cont = obj.frame().contents(((Repository)obj.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Provides));
        if (cont!=null && cont.length!=0) {
          for (int i=0; i<cont.length;i++) {
            InterfaceDef iface = SOFA.SOFAnode.Made.TIR.Access.TIRAccessMethods.getIfaceOfProvides((ProvideDef) cont[i]);
            String connName = cont[i].get_identification().name();
            if (SOFA.SOFAnode.Made.TIR.Access.TIRAccessMethods.containProvArrays((ProvideDef) cont[i]))
              connName = connName+"*";
            el = document.createElement("connector");
            el.setAttribute("name", connName);
         /*   el.setAttribute("unit", "Server");
            el.setAttribute("impl", "SGenerator:CSProcCallImpl"); */
            el.setAttribute("requires", contained2string(iface));  
            root.appendChild(el);
         /*   Element conEl = document.createElement("element");
            conEl.setAttribute("name", "sRole");
            conEl.setAttribute("impl", "EEM:SRoleImpl");
            el.appendChild(conEl);
            conEl = document.createElement("element");
            conEl.setAttribute("name", "skeleton");
            conEl.setAttribute("impl", "EEM:RMISkeleton");
            el.appendChild(conEl); */

            if (!isPrimitive) {
              el = document.createElement("connector");
              el.setAttribute("name", "_delegate_"+connName);
           /*   el.setAttribute("unit", "Client");
              el.setAttribute("impl", "SGenerator:CSProcCallImpl"); */
              el.setAttribute("provides", contained2string(iface));  
              root.appendChild(el);
           /*   conEl = document.createElement("element");
              conEl.setAttribute("name", "cRole");
              conEl.setAttribute("impl", "EEM:CRoleImpl");
              el.appendChild(conEl);
              conEl = document.createElement("element");
              conEl.setAttribute("name", "stub");
              conEl.setAttribute("impl", "EEM:RMIStub");
              el.appendChild(conEl);  */
            }
          }
        }

        cont = obj.frame().contents(((Repository)obj.get_containing_repository()).get_spec_def_kind(DefinitionKind.dk_Requires));
        if (cont!=null && cont.length!=0) {
          for (int i=0; i<cont.length;i++) {
            InterfaceDef iface = SOFA.SOFAnode.Made.TIR.Access.TIRAccessMethods.getIfaceOfRequires((RequireDef) cont[i]);
            String connName = cont[i].get_identification().name();
            if (SOFA.SOFAnode.Made.TIR.Access.TIRAccessMethods.containReqArrays((RequireDef) cont[i]))
              connName = connName+"*";
            el = document.createElement("connector");
            el.setAttribute("name", connName);
         /*   el.setAttribute("unit", "Client");
            el.setAttribute("impl", "SGenerator:CSProcCallImpl");*/
            el.setAttribute("provides", contained2string(iface));  
            root.appendChild(el);
         /*   Element conEl = document.createElement("element");
            conEl.setAttribute("name", "cRole");
            conEl.setAttribute("impl", "EEM:CRoleImpl");
            el.appendChild(conEl);
            conEl = document.createElement("element");
            conEl.setAttribute("name", "stub");
            conEl.setAttribute("impl", "EEM:RMIStub");  
            el.appendChild(conEl);  */

            if (!isPrimitive) {
              el = document.createElement("connector");
              el.setAttribute("name", "_subsume_"+connName);
           /*   el.setAttribute("unit", "Server");
              el.setAttribute("impl", "SGenerator:CSProcCallImpl");*/
              el.setAttribute("requires", contained2string(iface));  
              root.appendChild(el);
           /*   conEl = document.createElement("element");
              conEl.setAttribute("name", "sRole");
              conEl.setAttribute("impl", "EEM:SRoleImpl");
              el.appendChild(conEl);
              conEl = document.createElement("element");
              conEl.setAttribute("name", "skeleton");
              conEl.setAttribute("impl", "EEM:RMISkeleton");
              el.appendChild(conEl);  */
            }
          }
        }
      } catch (TIRExceptLock e) {
        throw new GenerException(Res.res.getString("MSG_LockedTIR"));
      }
      
      TransformerFactory tFactory = TransformerFactory.newInstance();
      Transformer transformer = tFactory.newTransformer();
//      transformer.setOutputProperty(javax.xml.transform.OutputKeys.INDENT, "yes");
      transformer.setOutputProperty("indent", "yes");
      DOMSource source = new DOMSource(document);
      CGFileWriter out = fs.newCGWriter(name+"CDD.xml");
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
      out.close();
    } catch (ParserConfigurationException e) {
      throw new GenerException(e.getMessage(), e);
    } catch (TransformerConfigurationException e) {
      throw new GenerException(e.getMessage(), e);
    } catch (TransformerException e) {
      throw new GenerException(e.getMessage(), e);
    }
  }
}
