/* $Id: JavaMapping.java,v 2.2 2002/06/20 10:10:18 hnetynka Exp $ */
package SOFA.SOFAnode.Made.CodeGen;

import SOFA.SOFAnode.Made.TIR.*;


/** Mapping of CDL types to Java.
  *
  * @author Petr Hnetynka
  */
public class JavaMapping implements Mapping {


  /** Returns name of the type in JAVA.
    * @param type    cdl type (TIR object)
    * @return   string with type name or null
    */
  public String getImplTypeName(CDLType type) {
    try {
      if (type instanceof PrimitiveDef) {
        int primKind = ((PrimitiveDef) type).kind().value();
        switch (primKind) {
          case PrimitiveKind.pk_void:
            return "void";
          case PrimitiveKind.pk_ushort:
            return "short";
          case PrimitiveKind.pk_short:
            return "short";
          case PrimitiveKind.pk_ulong:
            return "int";
          case PrimitiveKind.pk_long:
            return "int";
          case PrimitiveKind.pk_ulonglong:
            return "long";
          case PrimitiveKind.pk_longlong:
            return "long";
          case PrimitiveKind.pk_float:
            return "float";
          case PrimitiveKind.pk_double:
            return "double";
          case PrimitiveKind.pk_longdouble:
            return null;
          case PrimitiveKind.pk_boolean:
            return "boolean";
          case PrimitiveKind.pk_wchar:
            return "char";
          case PrimitiveKind.pk_char:
            return "char";
          case PrimitiveKind.pk_octet:
            return "byte";
          case PrimitiveKind.pk_wstring:
            return "java.lang.String";
          case PrimitiveKind.pk_string:
            return "java.lang.String";
          case PrimitiveKind.pk_any:
            return GeneratedObject.stdPackage+"Any";
          case PrimitiveKind.pk_object:
            return GeneratedObject.stdPackage+"Object";
        }
      } else {
        if (type instanceof TypedefDef) {
          int arr = 0;
          boolean ok = false;
          CDLType lastType = ((TypedefDef) type).original_type();
          while (!ok) {
            int dk = lastType.get_def_kind().value();
            switch (dk) {
              case DefinitionKind.dk_Array:
                arr++;
                lastType = ((ArrayDef) lastType).element_type();
                break;
              case DefinitionKind.dk_Sequence:
                arr++;
                lastType = ((SequenceDef) lastType).element_type();
                break;
              default:
                ok = true;
                break;
            }
          }
          if (arr == 0) {
            return getImplTypeName(lastType);
          } else {
            StringBuffer sb = new StringBuffer( getImplTypeName(lastType));
            sb.append(JavaGenerator.genBrackets(arr));
            return sb.toString();
          }
        } else {
          StringBuffer sb = new StringBuffer();
          Contained cnt = (Contained) type;
          while (! (cnt instanceof CDLRepository)) {
            Identification id = cnt.get_identification();
            String version = id.version();
            String name = id.name();
            if (GeneratedObject.isReservedName(name))
              name = "_"+name;
            if (sb.length() == 0) {
              sb.append(name);
            } else {
              sb.insert(0,".");
              if (cnt instanceof ModuleDef)
                sb.insert(0,name);
              else
                sb.insert(0,name+"Package");
            }
            cnt = (Contained) cnt.get_defined_in(); 
          }
          return sb.toString();
        }
      }
    } catch (java.rmi.RemoteException e) {
      return null;
    }
    return null;
  }


  /** Returns name of the holder in JAVA.
    * @param type    cdl type (TIR object)
    * @return   string with holder name or null
    */
  public String getImplTypeHolderName(CDLType type) {
    try {
      if (type instanceof PrimitiveDef) {
        int primKind = ((PrimitiveDef) type).kind().value();
        switch (primKind) {
          case PrimitiveKind.pk_void:
            return "null";
          case PrimitiveKind.pk_ushort:
            return GeneratedObject.stdPackage+"ShortHolder";
          case PrimitiveKind.pk_short:
            return GeneratedObject.stdPackage+"ShortHolder";
          case PrimitiveKind.pk_ulong:
            return GeneratedObject.stdPackage+"IntHolder";
          case PrimitiveKind.pk_long:
            return GeneratedObject.stdPackage+"IntHolder";
          case PrimitiveKind.pk_ulonglong:
            return GeneratedObject.stdPackage+"LongHolder";
          case PrimitiveKind.pk_longlong:
            return GeneratedObject.stdPackage+"LongHolder";
          case PrimitiveKind.pk_float:
            return GeneratedObject.stdPackage+"FloatHolder";
          case PrimitiveKind.pk_double:
            return GeneratedObject.stdPackage+"DoubleHolder";
          case PrimitiveKind.pk_longdouble:
            return null;
          case PrimitiveKind.pk_boolean:
            return GeneratedObject.stdPackage+"BooleanHolder";
          case PrimitiveKind.pk_wchar:
            return GeneratedObject.stdPackage+"CharHolder";
          case PrimitiveKind.pk_char:
            return GeneratedObject.stdPackage+"CharHolder";
          case PrimitiveKind.pk_octet:
            return GeneratedObject.stdPackage+"ByteHolder";
          case PrimitiveKind.pk_wstring:
            return GeneratedObject.stdPackage+"StringHolder";
          case PrimitiveKind.pk_string:
            return GeneratedObject.stdPackage+"StringHolder";
          case PrimitiveKind.pk_any:
            return GeneratedObject.stdPackage+"AnyHolder";
          case PrimitiveKind.pk_object:
            return GeneratedObject.stdPackage+"ObjectHolder";
        }
      } else {
        StringBuffer sb = new StringBuffer();
        Contained cnt = (Contained) type;
        while (! (cnt instanceof CDLRepository)) {
          Identification id = cnt.get_identification();
          String version = id.version();
          String name = id.name();
          if (GeneratedObject.isReservedName(name))
            name = "_"+name;
          if (sb.length() == 0) {
            sb.append(name);
          } else {
            sb.insert(0,".");
            if (cnt instanceof ModuleDef)
              sb.insert(0,name);
            else
              sb.insert(0,name+"Package");
          }
          cnt = (Contained) cnt.get_defined_in(); 
        }
        sb.append("Holder");
        return sb.toString();
        
      }
    } catch (java.rmi.RemoteException e) {
      return null;
    }
    return null;
  }

}
