/* $Id: PrimitiveGeneratedObject.java,v 2.1 2002/04/22 14:44:44 hnetynka Exp $ */
package SOFA.SOFAnode.Made.CodeGen;

import SOFA.SOFAnode.Made.TIR.PrimitiveKind;
import java.io.IOException;

/** For generating of references to primitive types (long, short, ...)
  * 
  * @author Petr Hnetynka
  */
public class PrimitiveGeneratedObject extends GeneratedObject {
  String inRef;
  String outRef;
  String defVal;
  String readVal;
  String writeVal;

  int primKind;

  /** Constructor.
    *
    * @param primKind kind of primitive type
    */
  public PrimitiveGeneratedObject(int primKind) {
    this.primKind = primKind;
    switch (primKind) {
     case PrimitiveKind.pk_void:
       inRef = "void";
       outRef = "";
       key = "void";
       defVal = "";
       readVal = "";
       writeVal = "";
       break;
     case PrimitiveKind.pk_ushort:
       inRef = "short";
       outRef = stdPackage+"ShortHolder";
       key = "ushort";
       defVal = "(short) 0";
       readVal = " = "+defaultIstream+".read_ushort()";
       writeVal = defaultOstream+".write_ushort(";
       break;
     case PrimitiveKind.pk_short:
       inRef = "short";
       outRef = stdPackage+"ShortHolder";
       key = "short";
       defVal = "(short) 0";
       readVal = " = "+defaultIstream+".read_short()";
       writeVal = defaultOstream+".write_short(";
       break;
     case PrimitiveKind.pk_ulong:
       inRef = "int";
       outRef = stdPackage+"IntHolder";
       key = "ulong";
       defVal = "(int) 0";
       readVal = " = "+defaultIstream+".read_ulong()";
       writeVal = defaultOstream+".write_ulong(";
       break;
     case PrimitiveKind.pk_long:
       inRef = "int";
       outRef = stdPackage+"IntHolder";
       key = "long";
       defVal = "(int) 0";
       readVal = " = "+defaultIstream+".read_long()";
       writeVal = defaultOstream+".write_long(";
       break;
     case PrimitiveKind.pk_ulonglong:
       inRef = "long";
       outRef = stdPackage+"LongHolder";
       key = "ulonglong";
       defVal = "(long) 0";
       readVal = " = "+defaultIstream+".read_ulonglong()";
       writeVal = defaultOstream+".write_ulonglong(";
       break;
     case PrimitiveKind.pk_longlong:
       inRef = "long";
       outRef = stdPackage+"LongHolder";
       key = "longlong";
       defVal = "(long) 0";
       readVal = " = "+defaultIstream+".read_longlong()";
       writeVal = defaultOstream+".write_longlong(";
       break;
     case PrimitiveKind.pk_float:
       inRef = "float";
       outRef = stdPackage+"FloatHolder";
       key = "float";
       defVal = "(float) 0";
       readVal = " = "+defaultIstream+".read_float()";
       writeVal = defaultOstream+".write_float(";
       break;
     case PrimitiveKind.pk_double:
       inRef = "double";
       outRef = stdPackage+"DoubleHolder";
       key = "double";
       defVal = "(double) 0";
       readVal = " = "+defaultIstream+".read_double()";
       writeVal = defaultOstream+".write_double(";
       break;
     case PrimitiveKind.pk_longdouble:
       key = "longdouble";
       inRef = "??? no mapping ???";
       outRef = "??? no mapping ???";
       break;  // !!!! no mapping
     case PrimitiveKind.pk_boolean:
       inRef = "boolean";
       outRef = stdPackage+"BooleanHolder";
       key = "boolean";
       defVal = "false";
       readVal = " = "+defaultIstream+".read_boolean()";
       writeVal = defaultOstream+".write_boolean(";
       break;
     case PrimitiveKind.pk_wchar:
       inRef = "char";
       outRef = stdPackage+"CharHolder";
       key = "wchar";
       defVal = "(char) 0";
       readVal = " = "+defaultIstream+".read_wchar()";
       writeVal = defaultOstream+".write_wchar(";
       break;
     case PrimitiveKind.pk_char:
       inRef = "char";
       outRef = stdPackage+"CharHolder";
       key = "char";
       defVal = "(char) 0";
       readVal = " = "+defaultIstream+".read_char()";
       writeVal = defaultOstream+".write_char(";
       break;
     case PrimitiveKind.pk_octet:
       inRef = "byte";
       outRef = stdPackage+"ByteHolder";
       key = "octet";
       defVal = "(byte) 0";
       readVal = " = "+defaultIstream+".read_octet()";
       writeVal = defaultOstream+".write_octet(";
       break;
     case PrimitiveKind.pk_wstring:
       inRef = "java.lang.String";
       outRef = stdPackage+"StringHolder";
       key = "wstring";
       defVal = "null";
       readVal = " = "+defaultIstream+".read_wstring()";
       writeVal = defaultOstream+".write_wstring(";
       break;
     case PrimitiveKind.pk_string:
       inRef = "java.lang.String";
       outRef = stdPackage+"StringHolder";
       key = "string";
       defVal = "null";
       readVal = " = "+defaultIstream+".read_string()";
       writeVal = defaultOstream+".write_string(";
       break;
     case PrimitiveKind.pk_any:
       inRef = stdPackage+"Any";
       outRef = stdPackage+"AnyHolder";
       key = "any";
       defVal = "null";
       readVal = " = "+defaultIstream+".read_any()";
       writeVal = defaultOstream+".write_any(";
       return;
     case PrimitiveKind.pk_object:
       inRef = stdPackage+"Object";
       outRef = stdPackage+"ObjectHolder";
       key = "object";
       defVal = "null";
       readVal = " = "+defaultIstream+".read_Object()";
       writeVal = defaultOstream+".write_Object(";
       return;
    }
  }

  // inherited from GeneratedObject
  public String genReference(int kind) {
    switch (kind) {
    case INREF:  return inRef;
    case OUTREF: return outRef;
    default:     return null;
    }
  }

  // inherited from GeneratedObject
  public void genReference(CGFileWriter out, int kind) throws IOException {
    switch (kind) {
    case INREF:  out.iwrite(inRef); break;
    case OUTREF: out.iwrite(outRef); break;
    default:     ;
    }
  }

  /** Creates generated objects for all primitive types
    * @param list   add objects to this list
    */
  public static void createPrimGenerObjects(GenObjectList list) {
    try {
      for (int i=0; i<18; i++) {
        list.addObject(new PrimitiveGeneratedObject(i));
      }
    } catch (GenObjectList.ExistException e) {
      Output.out.println("!!!!");
    }
  }

  public void defaultValue(CGFileWriter out) throws IOException {
    out.iwriteln(defVal);
  }

  public String defaultValue() throws IOException {
    return defVal;
  }

  public void genReadType(CGFileWriter out, String varName) throws IOException {
    out.iwrite(varName+readVal);
  }

  public String genReadType(String varName) throws IOException {
    return varName+readVal;
  }

  public void genWriteType(CGFileWriter out, String varName) throws IOException {
    out.iwrite(writeVal + varName + ")");
  }

  public String genWriteType(String varName) throws IOException {
    return writeVal + varName + ")";
  }

  public String toString() {
    return inRef;
  }
}
