/* $Id: ContainerImpl.java,v 2.0 2002/02/15 13:55:28 hnetynka Exp $ */
package SOFA.SOFAnode.Made.TIR.Impl;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;

public abstract class ContainerImpl extends UnicastRemoteObject implements Container, TIRImplObject, SContainer {
  protected int numOfItems;
  protected LiItem firstChild;
  protected LiItem lastChild;

  protected boolean hasCont;  /* for work object */

  public ContainerImpl() throws RemoteException {
    firstChild = lastChild = null;
    numOfItems = 0;
    hasCont = false;
  }

  /* add item to the list of childs */
  protected void addListItem(TIRObject what) {
    LiItem item = new LiItem(what);
    if (firstChild == null) {
      firstChild = lastChild = item;
    } else {
      LiItem akt = firstChild;
      while (akt.next != null) akt = akt.next;
      akt.next = item;
      item.prev = akt;
      lastChild = item;
    }
    numOfItems++;
  }

  public Contained[] contents(DefinitionKind type) throws RemoteException, TIRExceptLock {
    int i;
    Contained[] ret = new Contained [numOfItems];
    LiItem akt = firstChild;
    if ((type==null) || (type.value()==DefinitionKind.dk_all)) {
      for(i=0;i<numOfItems;i++) {
        ret[i] = (Contained) akt.obj;
        akt = akt.next;
      }
      return ret;
    }
    if (type.value()==DefinitionKind.dk_none) {
      return null;
    } else {
      int j = 0;
      for(i=0;i<numOfItems;i++) {
        if(akt.obj.get_def_kind().value() == type.value()) {
          ret[j] = (Contained) akt.obj;
	  j++;
	}
        akt = akt.next;
      }
      Contained[] r = new Contained [j];
      for(i=0;i<j;i++) {
        r[i] = ret[i];
      }
      return r;
    }
  }

  public Contained lookup(Identification id) throws RemoteException, TIRExceptLock {
    Contained ret = null;
    LiItem akt = firstChild;
    for(int i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_equal(id)) {
        ret = (Contained) akt.obj;
	break;
      } else {
        akt = akt.next;
      }
    }
    return ret;
  }

  public Contained   lookup(String name, String version) throws RemoteException, TIRExceptLock {
                                              /* !!! change cdl in whole file */
    Identification id = new IdentificationImpl("cdl","::"+name,version);
    Contained ret = null;
    LiItem akt = firstChild;
    for(int i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_equal(id)) {
        ret = (Contained) akt.obj;
	break;
      } else {
        akt = akt.next;
      }
    }
    return ret;
  }
  
  public Contained[] lookup_name(Identification id) throws RemoteException, TIRExceptLock {
    Contained[] all = new Contained [numOfItems];
    LiItem akt = firstChild;
    int i,j = 0;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_name_equal(id)) {
        all[j] = (Contained) akt.obj;
	j++;
      }
      akt = akt.next;
    }
    Contained[] ret = new Contained [j];
    for(i=0;i<j;i++) {
      ret[i] = all[i];
    }
    return ret;
  }

  public Contained[] lookup_name(String name) throws RemoteException, TIRExceptLock {
    Identification id = new IdentificationImpl("cdl","::"+name,"");
    Contained[] all = new Contained [numOfItems];
    LiItem akt = firstChild;
    int i,j = 0;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_name_equal(id)) {
        all[j] = (Contained) akt.obj;
	j++;
      }
      akt = akt.next;
    }
    Contained[] ret = new Contained [j];
    for(i=0;i<j;i++) {
      ret[i] = all[i];
    }
    return ret;   
  }

  public Contained lookup_tag(String name, String tag) throws RemoteException, TIRExceptLock {
    Identification idl;
    Contained ret = null;
    LiItem akt = firstChild;
    int i,j = 0;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) (idl = ((Contained)akt.obj).get_identification())).is_short_name_equal(name)) {
        String[] t = idl.tag();
        boolean found = false;
        for(int k=0;k<t.length;k++) {
          if (t[k].compareTo(tag)==0) {
            found = true;
            break;
          }
        }
        if (found) {
          ret = (Contained) akt.obj;
          break;
        }  
      }
      akt = akt.next;
    }
    return ret;   
  }
  public Contained[] lookup_branchtag(String name, String brtag) throws RemoteException, TIRExceptLock {
    Identification idl;
    Contained[] all = new Contained [numOfItems];
    LiItem akt = firstChild;
    int i,j = 0;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) (idl = ((Contained)akt.obj).get_identification())).is_short_name_equal(name)) {
        if (idl.branchTag().compareTo(brtag)==0) {
          all[j] = (Contained) akt.obj;
	  j++;
        }  
      }
      akt = akt.next;
    }
    Contained[] ret = new Contained [j];
    for(i=0;i<j;i++) {
      ret[i] = all[i];
    }
    return ret;
  }

  public Contained lookup_lastinbranch(String name, String brtag) throws RemoteException, TIRExceptLock {
    Identification idl;
    Contained ret = null;
    Contained[] branch = new Contained [numOfItems];
    LiItem akt = firstChild;
    if (brtag==null)
      brtag = "";
    int i,j = 0;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) (idl = ((Contained)akt.obj).get_identification())).is_short_name_equal(name)) {
        if (idl.branchTag().compareTo(brtag)==0) {
          branch[j] = (Contained) akt.obj;
	  j++;
        }  
      }
      akt = akt.next;
    }
    if (j!=0) {
      ret = branch[0];
      String ver = branch[0].get_identification().version();
      int last = ver.lastIndexOf('!');
      long max = 0;
      last++;
      while ((last < ver.length())&&(Character.isDigit(ver.charAt(last)))) {
        max *= 10;
        max += (int) (ver.charAt(last)-'0');
	last++;
      }
      for(i=1;i<j;i++) {
        ver = branch[i].get_identification().version();
        last = ver.lastIndexOf('!');
        long m = 0;
        last++;
        while ((last < ver.length())&&(Character.isDigit(ver.charAt(last)))) {
          m *= 10;
          m += (int) (ver.charAt(last)-'0');
	  last++;
        }

        if (m>max) {
          max = m;
          ret = branch[i];
        }
      }
    }
    return ret;
  }
  
  public Contained lookup_lastfromversion(String name, String version) throws RemoteException, TIRExceptLock {
    Contained tt = null;
    LiItem akt = firstChild;
    int i;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_equal(name,version)) {
        tt = (Contained) akt.obj;
	break;
      } else {
        akt = akt.next;
      }
    }
    if (tt==null) 
      return null;
     
    String brtag = tt.get_identification().branchTag();
    Identification idl;
    Contained ret = null;
    Contained[] branch = new Contained [numOfItems];
    akt = firstChild;
    int j = 0;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) (idl = ((Contained)akt.obj).get_identification())).is_short_name_equal(name)) {
        if (idl.branchTag().compareTo(brtag)==0) {
          branch[j] = (Contained) akt.obj;
	  j++;
        }  
      }
      akt = akt.next;
    }
    if (j!=0) {
      ret = branch[0];
      String ver = branch[0].get_identification().version();
      int last = ver.lastIndexOf('!');
      long max = 0;
      last++;
      while ((last < ver.length())&&(Character.isDigit(ver.charAt(last)))) {
        max *= 10;
        max += (int) (ver.charAt(last)-'0');
	last++;
      }
      for(i=1;i<j;i++) {
        ver = branch[i].get_identification().version();
        last = ver.lastIndexOf('!');
        long m = 0;
        last++;
        while ((last < ver.length())&&(Character.isDigit(ver.charAt(last)))) {
          m *= 10;
          m += (int) (ver.charAt(last)-'0');
	  last++;
        }

        if (m>max) {
          max = m;
          ret = branch[i];
        }
      }
    }
    return ret;
  }

  /* methods from iface SContainer */
  public Contained[] scontents(DefinitionKind type) throws RemoteException, TIRExceptLock {
    int i;
    Contained[] ret = new Contained [numOfItems];
    LiItem akt = firstChild;
    if ((type==null) || (type.value()==DefinitionKind.dk_all)) {
      for(i=0;i<numOfItems;i++) {
        ret[i] = (Contained) akt.obj;
        akt = akt.next;
      }
      return ret;
    }
    if (type.value()==DefinitionKind.dk_none) {
      return null;
    } else {
      int j = 0;
      for(i=0;i<numOfItems;i++) {
        if(akt.obj.get_def_kind().value() == type.value()) {
          ret[j] = (Contained) akt.obj;
	  j++;
	}
        akt = akt.next;
      }
      Contained[] r = new Contained [j];
      for(i=0;i<j;i++) {
        r[i] = ret[i];
      }
      return r;
    }
  }

  public Contained slookup(Identification id) throws RemoteException, TIRExceptLock {
    Contained ret = null;
    LiItem akt = firstChild;
    for(int i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_equal(id)) {
        ret = (Contained) akt.obj;
	break;
      } else {
        akt = akt.next;
      }
    }
    return ret;
  }

  public Contained slookup(String name, String version) throws RemoteException, TIRExceptLock {
    Identification id = new IdentificationImpl("cdl","::"+name,version);
    Contained ret = null;
    LiItem akt = firstChild;
    for(int i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_equal(id)) {
        ret = (Contained) akt.obj;
	break;
      } else {
        akt = akt.next;
      }
    }
    return ret;
  }
  
  public Contained[] slookup_name(Identification id) throws RemoteException, TIRExceptLock {
    Contained[] all = new Contained [numOfItems];
    LiItem akt = firstChild;
    int i,j = 0;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_name_equal(id)) {
        all[j] = (Contained) akt.obj;
	j++;
      }
      akt = akt.next;
    }
    Contained[] ret = new Contained [j];
    for(i=0;i<j;i++) {
      ret[i] = all[i];
    }
    return ret;
  }

  public Contained[] slookup_name(String name) throws RemoteException, TIRExceptLock {
    Identification id = new IdentificationImpl("cdl","::"+name,"");
    Contained[] all = new Contained [numOfItems];
    LiItem akt = firstChild;
    int i,j = 0;
    for(i=0;i<numOfItems;i++) {
      if (((SIdentification) ((Contained)akt.obj).get_identification()).is_short_name_equal(id)) {
        all[j] = (Contained) akt.obj;
	j++;
      }
      akt = akt.next;
    }
    Contained[] ret = new Contained [j];
    for(i=0;i<j;i++) {
      ret[i] = all[i];
    }
    return ret;   
  }

  public long isLockedForCreate()  throws RemoteException {
    return -1L;   /* modules (amodules, providers) and repository must this method rewrite */
  }
}


