/* $Id: InterfaceDefImpl.java,v 2.1 2002/04/17 13:31:06 hnetynka Exp $ */
package SOFA.SOFAnode.Made.TIR.Impl;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.*;

public class InterfaceDefImpl extends ContainerImpl implements InterfaceDef, SContained {
  protected Identification id;
  protected Container parent;
  protected Repository rep;
  DefinitionKindImpl defKindImpl;
  StateKindImpl stKindImpl;
  InterfaceDef[] base;
  CDLType[] loadBase;  // for loading
  ProtocolDef prot;

  InterfaceDefImpl normal;
  long workId;

  public InterfaceDefImpl(Identification ident, Container in, Repository inrep, InterfaceDef[] baseIf, long workId) throws RemoteException {
    id = ident;
    parent = in;
    rep = inrep;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Interface);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    if (baseIf==null)
      base = new InterfaceDef [0];
    else
      base = baseIf;
    prot = null;
    normal = null;
    this.workId = workId;     
  }

  public InterfaceDefImpl(Container in, Repository inrep) throws RemoteException {
    id = null;
    parent = in;
    rep = inrep;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Interface);
    stKindImpl = new StateKindImpl(StateKind.sk_normal);
    base = null;
    prot = null;
    normal = null;
    this.workId = -1L;
  }

  public InterfaceDefImpl(InterfaceDefImpl a, Container in, Repository inrep, long workId) throws RemoteException {
    id = a.id;
    parent = in;
    rep = inrep;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Interface);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    base = a.base;
    prot = a.prot;
    normal = a;
    this.workId = workId;
  }

  public InterfaceDef[] base_interfaces() throws RemoteException {
    return base;
  }

  public ProtocolDef protocol() throws RemoteException {
    return prot;
  }

  /* form interface Contained */
  public Identification get_identification() throws RemoteException {
    return id;
  }

  /* form interface Contained */
  public Container get_defined_in() throws RemoteException {
    return parent;
  }

  /* form interface Contained */
  public Container get_containing_repository() throws RemoteException {
    return rep;
  }

  /* form interface Contained */
  public AbsoluteName get_absolute_name() throws RemoteException {
    AbsoluteNameImpl absName;
    if (rep == parent) {  // object is contained in repository
      absName = new AbsoluteNameImpl("::"+id.name());
    } else {
      absName = new AbsoluteNameImpl(((SContained)parent).get_absolute_name().name()+"::"+id.name());
    }
    return (AbsoluteName) absName;
  }

  /* from interface TIRObject */
  public DefinitionKind get_def_kind() throws RemoteException {
    return (DefinitionKind) defKindImpl;
  }

  /* from interface TIRObject */
  public StateKind get_state() throws RemoteException {
    return (StateKind) stKindImpl;
  }

  public void tag(String t) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (stKindImpl.value()!=StateKind.sk_work)
      throw new TIRExceptCreate("you can call create method on work object");
    String ebt = id.branchTag();
    if (t.compareTo(ebt)==0)
      throw new TIRExceptCreate("This tag exists");
    String[] et = id.tag();
    int i,j;
    for (i=0;i<et.length;i++) {
      if (t.compareTo(et[i])==0)
       throw new TIRExceptCreate("This tag exists");       
    }
    Contained[] sibl = parent.lookup_name(id);
    for (j=0;j<sibl.length;j++) {
      Identification idl = sibl[j].get_identification();
      
      ebt = idl.branchTag();
      if (t.compareTo(ebt)==0)
        throw new TIRExceptCreate("This tag exists");
      et = idl.tag();
      for (i=0;i<et.length;i++) {
        if (t.compareTo(et[i])==0)
          throw new TIRExceptCreate("This tag exists");       
      } 
    }
    ((SIdentification) id).tag(t);
  }

  private void fromNormObj() throws RemoteException, TIRExceptLock {
    if (normal==null)
      return;
    if (((WorkRepositoryImpl)rep).lock.isLocked())
      throw new TIRExceptLock("Repository is locked");
    int i;
    LiItem akt;
    akt = normal.firstChild;
    for(i=0;i<normal.numOfItems;i++) {
      addListItem(WorkRepositoryImpl.newWorkFromNormal(akt.obj,rep,this,workId));
      akt = akt.next;
    }
    hasCont = true;
  }

  public Contained[] contents(DefinitionKind type) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.contents(type);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.contents(type);     
    }
  }    

  public Contained lookup(Identification id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup(id);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup(id);
    }
  }

  public Contained lookup(String name, String ver) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup(name, ver);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup(name, ver);
    }
  }

  public Contained[] lookup_name(Identification id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_name(id);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_name(id);
    }
  }

  public Contained[] lookup_name(String id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_name(id);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_name(id);
    }
  }

  public Contained lookup_tag(String name, String tag) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_tag(name, tag);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_tag(name, tag);
    }
  }

  public Contained[] lookup_branchtag(String name, String brtag) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_branchtag(name, brtag);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_branchtag(name, brtag);
    }
  }

  public Contained lookup_lastinbranch(String name, String brtag) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_lastinbranch(name, brtag);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_lastinbranch(name, brtag);
    }
  }
  
  public Contained lookup_lastfromversion(String name, String version) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_lastfromversion(name, version);
    } else {
      if (!hasCont)
        fromNormObj();
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      else 
        return super.lookup_lastfromversion(name, version);
    }
  }

  public Contained[] scontents(DefinitionKind type) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      return super.scontents(type);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.scontents(type);
    }
  }

  public Contained slookup(Identification id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
        return super.slookup(id);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.slookup(id);
    }
  }

  public Contained slookup(String name, String version) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
        return super.slookup(name, version);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.slookup(name, version);
    }
  }

  public Contained[] slookup_name(Identification id) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
        return super.slookup_name(id);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.slookup_name(id);
    }
  }

  public Contained[] slookup_name(String name) throws RemoteException, TIRExceptLock {
    if (stKindImpl.value()==StateKind.sk_normal) {
      return super.slookup_name(name);
    } else {
      if (!hasCont)
        fromNormObj();
      return super.slookup_name(name);
    }
  }

  public StructDef create_struct(Identification id) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id!=null) {
      if (!hasCont)
        fromNormObj();
      // set version of substructure same as this version  
      id = new IdentificationImpl(((SIdentification)id).language(), id.absolute_name().name(),this.id.version());
      Contained[] a = lookup_name(id.name());
      if (a.length != 0) {
        if (a[0].get_def_kind().value() != DefinitionKind.dk_Struct)
	  throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	boolean found = false;  
	for (int i=0;i<a.length;i++) {
	  if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	    found = true;
	    break;
	  }
	}
	if (found) 
	  throw new TIRExceptCreate("Object with same identification exists in repository.");
      }
      // creating new object
      StructDef ret = new StructDefImpl(id, this, rep,workId);
      addListItem(ret); // adding object to list
      return ret;       // return object
    } else
      return null;
  }
  
  public UnionDef create_union(Identification id, CDLType switch_type) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id==null || switch_type==null)
      return null;
    else {
      if (switch_type.get_state().value()==StateKind.sk_work ) { 
        if (!hasCont)
          fromNormObj();
        // set version of substructure same as this version  
        id = new IdentificationImpl(((SIdentification)id).language(),  id.absolute_name().name(),this.id.version());
        Contained[] a = lookup_name(id.name());
        if (a.length != 0) {
          if (a[0].get_def_kind().value() != DefinitionKind.dk_Union)
	    throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	  boolean found = false;  
	  for (int i=0;i<a.length;i++) {
	    if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	      found = true;
	      break;
	    }
	  }
	  if (found) 
	    throw new TIRExceptCreate("Object with same identification exists in repository.");
        }
        // creating new object
        UnionDef ret = new UnionDefImpl(id, this, rep, switch_type,workId);
        addListItem(ret); // adding object to list
        return ret;       // return object
      } else {
        throw new TIRExceptCreate("Given object isn't work.");
      }
    }
  }
  
  public EnumDef create_enum(Identification id) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id!=null) {
      if (!hasCont)
        fromNormObj();
      // set version of substructure same as this version  
      id = new IdentificationImpl(((SIdentification)id).language(), id.absolute_name().name(),this.id.version());
      Contained[] a = lookup_name(id.name());
      if (a.length != 0) {
        if (a[0].get_def_kind().value() != DefinitionKind.dk_Enum)
	  throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	boolean found = false;  
	for (int i=0;i<a.length;i++) {
	  if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	    found = true;
	    break;
	  }
	}
	if (found) 
	  throw new TIRExceptCreate("Object with same identification exists in repository.");
      }
      // creating new object
      EnumDef ret = new EnumDefImpl(id, this, rep, workId);
      addListItem(ret); // adding object to list
      return ret;       // return object
    } else
      return null;
  }

  public AttributeDef create_attribute(Identification name, CDLType type, int mode) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (type==null)
      return null;
    if (type.get_state().value()==StateKind.sk_work) {
      // set version of substructure same as this version  
      Identification id = new IdentificationImpl(((SIdentification)name).language(), name.absolute_name().name(),this.id.version());
      Contained[] a = lookup_name(name.name());
      if (a.length != 0)
	throw new TIRExceptCreate("Name \""+name.name()+"\" exists in repository.");
      // creating new object
      AttributeDef ret = new AttributeDefImpl(id, this, rep, type, new AttrModeImpl(mode), workId);
      addListItem(ret); // adding object to list
      return ret;       // return object
    } else {
      throw new TIRExceptCreate("Given object isn't work.");
    }
  }

  public OperationDef create_operation(Identification name, CDLType result, ExceptionDef[] exceptions) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (result==null || exceptions==null)
      return null;
    int i;
    for (i=0;i<exceptions.length;i++)
      if (exceptions[i]==null)
        return null;
    if (result.get_state().value()==StateKind.sk_normal) 
      throw new TIRExceptCreate("Given object isn't work.");
    for (i=0;i<exceptions.length;i++)
      if (exceptions[i].get_state().value()==StateKind.sk_normal)
        throw new TIRExceptCreate("Given object isn't work.");
    // set version of substructure same as this version  
    Identification id = new IdentificationImpl(((SIdentification)name).language(), name.absolute_name().name(),this.id.version());
    Contained[] a = lookup_name(name.name());
    if (a.length != 0)
       throw new TIRExceptCreate("Name \""+name.name()+"\" exists in repository.");
      // creating new object
    ExceptionDef[] ne = new ExceptionDef [exceptions.length];
    for (i=0;i<exceptions.length;i++)
      ne[i]=exceptions[i];
    OperationDef ret = new OperationDefImpl(id, this, rep, result, ne, workId);
    addListItem(ret); // adding object to list
    return ret;       // return object
  }

  public ProtocolDef create_protocol() throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("You can call create method on work object only");
    if (prot!=null)
      throw new TIRExceptCreate("Protocol already exists");
    prot = new ProtocolDefImpl(StateKind.sk_work);
    return prot;
  }

  public ConstantDef create_constant(Identification id, CDLType type, ExprOperDef value) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id==null || type==null || value==null)
      return null;
    else {
      if (type.get_state().value()==StateKind.sk_work && 
          value.get_state().value()==StateKind.sk_work) {
        if (!hasCont)
          fromNormObj();
        // set version of substructure same as this version  
        id = new IdentificationImpl(((SIdentification)id).language(), id.absolute_name().name(),this.id.version());
        Contained[] a = lookup_name(id.name());
        if (a.length != 0) {
          if (a[0].get_def_kind().value() != DefinitionKind.dk_Constant)
	    throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	  boolean found = false;  
	  for (int i=0;i<a.length;i++) {
	    if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	      found = true;
	      break;
	    }
	  }
	  if (found) 
	    throw new TIRExceptCreate("Object with same identification exists in repository.");
        }
        // creating new object
        ConstantDef ret = new ConstantDefImpl(id, this, rep,type, value, workId);
        addListItem(ret); // adding object to list
        return ret;       // return object
      } else {
        throw new TIRExceptCreate("Given objects aren't work.");
      }
    }
  }

  public TypedefDef create_typedef(Identification id, CDLType orig_type) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id==null || orig_type==null)
      return null;
    else {
      if (orig_type.get_state().value()==StateKind.sk_work ) { 
        if (!hasCont)
          fromNormObj();
        // set version of substructure same as this version  
        id = new IdentificationImpl(((SIdentification)id).language(), id.absolute_name().name(),this.id.version());
        Contained[] a = lookup_name(id.name());
        if (a.length != 0) {
          if (a[0].get_def_kind().value() != DefinitionKind.dk_Typedef)
	    throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	  boolean found = false;  
	  for (int i=0;i<a.length;i++) {
	    if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	      found = true;
	      break;
	    }
	  }
	  if (found) 
	    throw new TIRExceptCreate("Object with same identification exists in repository.");
        }
        // creating new object
        TypedefDef ret = new TypedefDefImpl(id, this, rep, orig_type,workId);
        addListItem(ret); // adding object to list
        return ret;       // return object
      } else {
        throw new TIRExceptCreate("Given object isn't work.");
      }
    }
  }
  
  public ExceptionDef create_exception(Identification id) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (!isNew())
      throw new TIRExceptCreate("you can call create method on work object");
    if (id!=null) {
      if (!hasCont)
        fromNormObj();
      // set version of substructure same as this version  
      id = new IdentificationImpl(((SIdentification)id).language(), id.absolute_name().name(),this.id.version());
      Contained[] a = lookup_name(id.name());
      if (a.length != 0) {
        if (a[0].get_def_kind().value() != DefinitionKind.dk_Exception)
	  throw new TIRExceptCreate("Name \""+id.name()+"\" exists in repository.");
	boolean found = false;  
	for (int i=0;i<a.length;i++) {
	  if (((SIdentification) a[i].get_identification()).is_short_equal(id)) {
	    found = true;
	    break;
	  }
	}
	if (found) 
	  throw new TIRExceptCreate("Object with same identification exists in repository.");
      }
      // creating new object
      ExceptionDef ret = new ExceptionDefImpl(id, this, rep,workId);
      addListItem(ret); // adding object to list
      return ret;       // return object
    } else
      return null;
  }


  public void save(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      int i;
      st.curOutFile = new DataOutputStream(new FileOutputStream(st.currentFile));
      ((IdentificationImpl)id).save(st.curOutFile);   // saving identification
      // saving protocol
      if (prot != null) {
        st.curOutFile.writeBoolean(true);
        ((TIRImplObject) prot).save(st);
      } else {
         st.curOutFile.writeBoolean(false);       
      }
      // --- and of saving protocol
      st.curOutFile.writeInt(base.length);
      for(i=0;i<base.length;i++) {
        Storage.writeCDLType(st, base[i]);
      }
      st.curOutFile.writeInt(numOfItems);
      LiItem akt = firstChild;
      for(i=0;i<numOfItems;i++) {  // saving all childs
        st.curOutFile.writeInt(akt.obj.get_def_kind().value());
        ((TIRImplObject)(akt.obj)).save(st);
	akt = akt.next;
      }
      st.curOutFile.close();
    } catch (IOException e) {
      throw new TIRExceptStorage("Access error in "+st.currentFile+".");
    }
  }

  /*from interface TIRImplObject */
  public void load(Storage st) throws RemoteException, TIRExceptStorage{
    try {
      int num,i;
      st.curInFile = new DataInputStream(new FileInputStream(st.currentFile));
      id = new IdentificationImpl();
      ((IdentificationImpl)id).load(st.curInFile);
      // loading protocol
      if (st.curInFile.readBoolean()) {
        prot = new ProtocolDefImpl(StateKind.sk_normal);
        ((TIRImplObject) prot).load(st);
      } else {
        prot = null;
      }
      // --- end of loading protocol
      num = st.curInFile.readInt();
      loadBase = new CDLType [num];
      for(i=0;i<num;i++) {
        loadBase[i] = Storage.readCDLType(st);
      }
      num = st.curInFile.readInt();
      int dkind;
      TIRImplObject obj;
      for (i=0;i<num;i++) {
        dkind = st.curInFile.readInt();
	switch (dkind) {
          case DefinitionKind.dk_Struct:
	    obj = new StructDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Enum:
	    obj = new EnumDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Union:
	    obj = new UnionDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Exception:
	    obj = new ExceptionDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Constant:
	    obj = new ConstantDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Typedef:
	    obj = new TypedefDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Attribute:
	    obj = new AttributeDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  case DefinitionKind.dk_Operation:
	    obj = new OperationDefImpl(this, rep);
	    obj.load(st);
	    addListItem((TIRObject)obj);
	    break;
	  default:
            throw new TIRExceptStorage("Unexpected kind of object in \""
	                              +st.currentFile+"\".");
	} // -- end of switch
      }
      st.curInFile.close();
    } catch (IOException e) {
      throw new TIRExceptStorage("Access error in "+st.current+".");
    }
  }

  public void postLoad(RepositoryImpl r) throws RemoteException, TIRExceptStorage {
    try {
  
    int i,k;

    if (prot != null)
      ((TIRImplObject) prot).postLoad(r);

    LiItem akt = firstChild;
    for(i=0;i<numOfItems;i++) {
      ((TIRImplObject) akt.obj).postLoad(r);
      akt = akt.next;
    }

    // postLoad base
    Container con;
    Contained[] rt;
    Contained ret;
    IdentificationImpl id;
    AbsoluteName nm;
    boolean found;
    base = new InterfaceDef [loadBase.length];
    for (int j=0;j<loadBase.length;j++) {
      if (loadBase[j].get_def_kind().value() == DefinitionKind.dk_none) {   // is tp NotYetInRep?
        nm = ((NotYetInRep) loadBase[j]).nm;
  	con = r;  // begin search from repository
        for (i=0;i<nm.size()-1;i++) {  // last element is name of object
  	  id = new IdentificationImpl(nm.elementAt(0),"::"+nm.elementAt(i),"");
          rt = ((SContainer)con).slookup_name(id);
  	  if (rt.length==0) throw new TIRExceptStorage("Can't find interface " + nm.name());
            if (rt.length==1) {
              con = (Container) rt[0];
  	  } else { 
  	    found = false;
            id.version(((NotYetInRep) loadBase[j]).id.version());
  	    for (k=0;k<rt.length;k++) {
  	      if (id.is_short_equal(rt[k].get_identification())) {
  	        found = true;
  		break;  // -- break for
  	      }
  	    } // -- end of for
  	    if (!found) throw new TIRExceptStorage("Can't find interface " + nm.name());
  	    con = (Container) rt[k];
  	  }
  	}
  	if ((ret = ((SContainer)con).slookup(((NotYetInRep) loadBase[j]).id))==null)
  	  throw new TIRExceptStorage("Can't find interface " + nm.name());
  	base[j] = (InterfaceDef) ret;
	if (base[j].get_def_kind().value()==DefinitionKind.dk_Array ||
            base[j].get_def_kind().value()==DefinitionKind.dk_Sequence)
	  ((TIRImplObject) base[j]).postLoad(r);
      }
    }

    } catch (TIRExceptLock e) {;}
  }

 public boolean isNew() {
    return ((stKindImpl.value()==StateKind.sk_work) && normal==null);
  }

  public void canCommit() throws RemoteException, TIRExceptCommit {
    if (stKindImpl.value()==StateKind.sk_normal)
      return;
    LiItem akt;
    int i;
    akt = firstChild;
    for (i=0;i<numOfItems;i++) {
      ((TIRImplObject)akt.obj).canCommit();
      akt = akt.next;
    }
    if (prot != null)
      ((TIRImplObject)prot).canCommit();
  }

  public void doCommit(Container in, Repository rep) throws RemoteException {
    if (stKindImpl.value()==StateKind.sk_normal)
      return;
    if (isNew()) {  
      int i;
      stKindImpl.toNormal();
      LiItem akt = firstChild;
      for (i=0;i<numOfItems;i++) {
        ((TIRImplObject)akt.obj).doCommit(this,rep);
        akt = akt.next;     
      }
      loadBase = new CDLType [base.length];
      for (i=0;i<base.length;i++) {
        Identification idl = ((Contained)base[i]).get_identification();
        loadBase[i] = new NotYetInRep(idl.lang_absolute_name().name(), idl.name(), idl.version());
      }
      if (prot != null)
        ((TIRImplObject)prot).doCommit(this, rep);
      base = null;
      parent = in;
      this.rep = rep;
    } else {
      normal = null;
    }
  }

  public void doAbort(long workId) throws RemoteException {
    LiItem akt = firstChild;
    for(int i=0;i<numOfItems;i++) {
      if (((TIRImplObject)akt.obj).isNew()) 
        ((TIRImplObject)akt.obj).doAbort(workId);
      akt = akt.next;
    }  
  }

  public void setBaseInterfaces(InterfaceDef[] base_interfaces) throws RemoteException, TIRExceptCreate {
    if (base_interfaces == null)
      base = new InterfaceDef [0];
    else {
      base = new InterfaceDef [base_interfaces.length];
      for (int i=0;i<base_interfaces.length;i++) {
        if (base_interfaces[i].get_state().value()==StateKind.sk_work)
	  base[i]=base_interfaces[i];
	else
	  throw new TIRExceptCreate("Given objects aren't work.");
      }
    }
  }
}
