/* $Id: ProfileDefImpl.java,v 2.0 2002/02/15 13:55:34 hnetynka Exp $ */
package SOFA.SOFAnode.Made.TIR.Impl;
import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.*;

public class ProfileDefImpl extends UnicastRemoteObject implements ProfileDef, TIRImplObject, SProfileDef {
  DefinitionKindImpl defKindImpl;
  StateKindImpl stKindImpl;

  String nm;
  LiIdItem firstId;
  LiIdItem lastId;
  int numOfIds;
 
  Repository rep;
 
  boolean isNewP;
  boolean hasCont;
  ProfileDefImpl normal;
  boolean changed;

  Lock plock;
  long lockFor;      // in normal
  
  long contentId;    // in work and normal, if normal > work -> reload content to work
  boolean lockForMe;  // in work

  public ProfileDefImpl(String name, Repository r, int state) throws RemoteException {
    nm = name;
    firstId = lastId = null;
    numOfIds = 0;
    plock = new Lock(false);
    lockFor = -1L;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Profile);
    stKindImpl = new StateKindImpl(state);
    if (state==StateKind.sk_work)
      isNewP = true;
    else
      isNewP = false;
    hasCont = true;
    normal = null;
    rep = r;
    changed = true;
    contentId = 0;
  }

  public ProfileDefImpl(Repository r) throws RemoteException {
    nm = null;
    firstId = lastId = null;
    numOfIds = 0;
    plock = new Lock(false);
    lockFor = -1L;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Profile);
    stKindImpl = new StateKindImpl(StateKind.sk_normal);
    isNewP = false;
    hasCont = true;
    normal = null;
    rep = r;
    changed = true;
    contentId = 0;
  }

  public ProfileDefImpl(ProfileDefImpl a, Repository r) throws RemoteException {
    nm = a.nm;
    firstId = lastId = null;
    numOfIds = 0;
    plock = a.plock;
    lockFor = -1L;
    defKindImpl = new DefinitionKindImpl(DefinitionKind.dk_Profile);
    stKindImpl = new StateKindImpl(StateKind.sk_work);
    isNewP = false;
    hasCont = false;
    normal = a;
    rep = r;
    changed = false;
    contentId = 0;
  }

  public DefinitionKind get_def_kind() throws RemoteException {
    return (DefinitionKind) defKindImpl;
  }

  public StateKind get_state() throws RemoteException {
    return (StateKind) stKindImpl;
  }

  public String name() throws RemoteException {
    return nm;
  }
  
  public Identification[] contents() throws RemoteException, TIRExceptLock {
    int i;
    if (stKindImpl.value() == StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      Identification[] ret = new Identification [numOfIds];
      LiIdItem akt = firstId;
      for(i=0;i<numOfIds;i++) {
        ret[i] = akt.obj;
        akt = akt.next;
      }
      return ret;
    } else {
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      if (normal!=null) {
        if ((!hasCont) || (normal.contentId > this.contentId)) {
          firstId = lastId = null;
          numOfIds = 0;
          LiIdItem akt = normal.firstId;
          for(i=0;i<normal.numOfIds;i++) {
            LiIdItem item = new LiIdItem(akt.obj);
            if (firstId == null) {
              firstId = lastId = item;
            } else {
              lastId.next = item;
              item.prev = lastId;
              lastId = item;
            }
            numOfIds++;
            
            akt = akt.next;
          }
          this.contentId = normal.contentId;
        }
        hasCont=true;
      }
      Identification[] ret = new Identification [numOfIds];
      LiIdItem akt = firstId;
      for(i=0;i<numOfIds;i++) {
        ret[i] = akt.obj;
        akt = akt.next;
      }
      return ret;
    }
  }
  
  public Identification lookup(String fullName) throws RemoteException, TIRExceptLock {
    int i;
    fullName = fullName.trim();
    if (stKindImpl.value() == StateKind.sk_normal) {
      if (((RepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      //AbsoluteName an = new AbsoluteNameImpl(fullName);
      LiIdItem akt = firstId;
      //AbsoluteName akan;
      for (i=0;i<numOfIds;i++) {
        //akan = akt.obj.lang_absolute_name();
        //if (akan.isEqual(an))
        if (akt.key.compareTo(fullName)==0)
          return akt.obj;
        akt = akt.next;  
      }
      return null;
    } else {
      if (((WorkRepositoryImpl) rep).lock.isLocked())
        throw new TIRExceptLock("Repository is locked");
      if (normal!=null) {
        if ((!hasCont) || (normal.contentId > this.contentId)) {
          firstId = lastId = null;
          numOfIds = 0;
          LiIdItem akt = normal.firstId;
          for(i=0;i<normal.numOfIds;i++) {
            LiIdItem item = new LiIdItem(akt.obj);
            if (firstId == null) {
              firstId = lastId = item;
            } else {
              lastId.next = item;
              item.prev = lastId;
              lastId = item;
            }
            numOfIds++;
            
            akt = akt.next;
          }
          contentId = normal.contentId;
        }
        hasCont=true;
      }
      //AbsoluteName an = new AbsoluteNameImpl(fullName);
      LiIdItem akt = firstId;
      //AbsoluteName akan;
      for (i=0;i<numOfIds;i++) {
        //akan = akt.obj.lang_absolute_name();
        //if (akan.isEqual(an))
        if (akt.key.compareTo(fullName)==0)
          return akt.obj;
        akt = akt.next;  
      }
      return null;
    }
  }
  
  public boolean add(Identification a) throws RemoteException, TIRExceptLock, TIRExceptCreate {
    if (stKindImpl.value() == StateKind.sk_normal)
      throw new TIRExceptCreate("You can call create only on work object");

    if (((WorkRepositoryImpl) rep).lock.isLocked())
      throw new TIRExceptLock("Repository is locked");
    if (normal!=null) {
      if ((!hasCont) || (normal.contentId > this.contentId)) {
        firstId = lastId = null;
        numOfIds = 0;
        LiIdItem akt = normal.firstId;
        for(int i=0;i<normal.numOfIds;i++) {
          LiIdItem item = new LiIdItem(akt.obj);
          if (firstId == null) {
            firstId = lastId = item;
          } else {
            lastId.next = item;
            item.prev = lastId;
            lastId = item;
          }
          numOfIds++;
          akt = akt.next;
        }
        contentId = normal.contentId;
      }
      hasCont=true;
    }

    ((WorkRepositoryImpl) rep).normalRep.lockProfiles.lock();
    if (this.isLocked()) {
      if (!lockForMe) {
        ((WorkRepositoryImpl) rep).normalRep.lockProfiles.unlock();
        throw new TIRExceptLock("Profile is locked.");
      }
    }
    if (!this.lock()) {
      ((WorkRepositoryImpl) rep).normalRep.lockProfiles.unlock();
      throw new TIRExceptLock("Profile can not lock.");
    }  
    ((WorkRepositoryImpl) rep).normalRep.lockProfiles.unlock();
    
    //AbsoluteName an = a.lang_absolute_name();
    String ans = a.lang_absolute_name().name();
    //AbsoluteName akan;
    LiIdItem akt = firstId;
    boolean found = false;
    for(int i=0;i<numOfIds;i++) {
      //akan = akt.obj.lang_absolute_name();
      //if (akan.isEqual(an))
      if (akt.key.compareTo(ans)==0)
        return false;
      akt = akt.next;
    }
    
    LiIdItem item = new LiIdItem(a);
    if (firstId == null) {
       firstId = lastId = item;
    } else {
      lastId.next = item;
      item.prev = lastId;
      lastId = item;
    }
    numOfIds++;
    changed = true;
    return true;
  }
  
  public void remove(String fullName) throws RemoteException, TIRExceptCreate, TIRExceptLock {
    if (stKindImpl.value() == StateKind.sk_normal)
      throw new TIRExceptCreate("You can call create only on work object");

    if (((WorkRepositoryImpl) rep).lock.isLocked())
      throw new TIRExceptLock("Repository is locked");
    if (normal!=null) {
      if ((!hasCont) || (normal.contentId > this.contentId)) {
        firstId = lastId = null;
        numOfIds = 0;
        LiIdItem akt = normal.firstId;
        for(int i=0;i<normal.numOfIds;i++) {
          LiIdItem item = new LiIdItem(akt.obj);
          if (firstId == null) {
            firstId = lastId = item;
          } else {
            lastId.next = item;
            item.prev = lastId;
            lastId = item;
          }
          numOfIds++;
          akt = akt.next;
        }
        contentId = normal.contentId;
      }
      hasCont=true;
    }

    ((WorkRepositoryImpl) rep).normalRep.lockProfiles.lock();
    if (this.isLocked()) {
      if (!lockForMe) {
        ((WorkRepositoryImpl) rep).normalRep.lockProfiles.unlock();
        throw new TIRExceptLock("Profile is locked.");
      }
    }
    if (!this.lock()) {
      ((WorkRepositoryImpl) rep).normalRep.lockProfiles.unlock();
      throw new TIRExceptLock("Profile can not lock.");
    }  
    ((WorkRepositoryImpl) rep).normalRep.lockProfiles.unlock();

    AbsoluteName an = new AbsoluteNameImpl(fullName);
    LiIdItem akt = firstId;
    AbsoluteName akan;
    for (int i=0;i<numOfIds;i++) {
      akan = akt.obj.lang_absolute_name();
      if (akan.isEqual(an)) {
        if (numOfIds==1) {
          firstId = lastId = null;
          numOfIds = 0;
        } else {
          if (akt == firstId) {
            firstId = akt.next;
  	    firstId.prev = null;
	    akt.next = null;
	  } else {
            if (lastId == akt) {
  	      lastId = akt.prev;
              lastId.next = null;
              akt.prev = null;
	    } else {
	      akt.prev.next = akt.next;
              akt.next.prev = akt.prev;
	    }
	  }
          numOfIds--;
        }
        changed = true;
        return;
      }  
      akt = akt.next;  
    }
  }

  public boolean isLocked() throws RemoteException {
    return plock.isLocked();
  }
  
  public boolean lock() throws RemoteException {
    if (this.lockForMe)
      return true;
    boolean a = plock.tryLock();
    if (a) {
      if (normal!=null)
        normal.lockFor = ((WorkRepositoryImpl) rep).workId;
      this.lockForMe = true;
    }
    return a;
  }
  
  public void unlock() throws RemoteException {
    lockForMe = false;
    if (normal!=null)
      normal.lockFor = -1L;
    plock.unlock();
  }

  public void save(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      st.curOutFile = new DataOutputStream(new FileOutputStream(st.currentFile));
      Storage.writeString(st.curOutFile,nm);
      LiIdItem akt = firstId;
      st.curOutFile.writeInt(numOfIds);
      for (int i=0;i<numOfIds;i++) {
        Storage.writeString(st.curOutFile,akt.obj.lang_absolute_name().name());
        Storage.writeString(st.curOutFile,akt.obj.version());
        akt = akt.next;
      }
      st.curOutFile.close();
    } catch (IOException e) {
      throw new TIRExceptStorage("Access error in "+st.current+".");
    }
  }

  public void load(Storage st) throws RemoteException, TIRExceptStorage {
    try {
      st.curInFile = new DataInputStream(new FileInputStream(st.currentFile));
      nm = Storage.readString(st.curInFile);
      int num = st.curInFile.readInt();
      for (int j=0;j<num;j++) {
        String absName = Storage.readString(st.curInFile);
        String version = Storage.readString(st.curInFile);
        try {
          AbsoluteName nm = new AbsoluteNameImpl(absName);
       	  Container con = rep;  // begin search from repository
    	  Contained[] rt;
    	  Contained ret;
    	  IdentificationImpl id;
          for (int i=0;i<nm.size()-1;i++) {  // last element is name of object
            id = new IdentificationImpl(nm.elementAt(0), "::"+nm.elementAt(i),"");
            rt = ((SContainer) con).slookup_name(id);
    	    if (rt.length==0) throw new TIRExceptStorage("Can't find type " + nm.name());
            if (rt.length==1) {
              con = (Container) rt[0];
    	    } else {  // type in struct (etc.) - more versions of this struct
    	      boolean found = false;
              id.version(version);  // substruct version is inherited from parent
              int k = 0;
    	      for (k=0;k<rt.length;k++) {
    	        if (id.is_short_equal(rt[k].get_identification())) {
    	          found = true;
                  break;  // -- break for
    	        }
    	      } // -- end of for
    	      if (!found) throw new TIRExceptStorage("Can't find type " + nm.name());
    	      con = (Container) rt[k];
    	    }
    	  }
    	  if ((ret = ((SContainer)con).slookup(nm.elementAt(nm.size()-1),version))==null)
    	    throw new TIRExceptStorage("Can't find type " + nm.name());
    	  LiIdItem item = new LiIdItem(ret.get_identification());

          if (firstId == null) {
            firstId = lastId = item;
          } else {
            lastId.next = item;
            item.prev = lastId;
            lastId = item;
          }
          numOfIds++;
        } catch (TIRExceptLock e) {;}
      }
      
      st.curInFile.close();
    } catch (IOException e) {
      throw new TIRExceptStorage("Access error in "+st.current+".");
    }
  }

  public void postLoad(RepositoryImpl r) throws RemoteException, TIRExceptStorage {}
  public boolean isNew() { return isNewP; }
  public void canCommit() throws RemoteException, TIRExceptCommit {
    // testing if all identification in this profile is OK
    try {
    LiIdItem akt = firstId;
    for (int k=0;k<numOfIds;k++) {
      AbsoluteName nm = akt.obj.lang_absolute_name();
      Container con = rep;  // begin search from repository
      Contained[] rt;
      Contained ret;
      IdentificationImpl id;
      for (int i=0;i<nm.size()-1;i++) {  // last element is name of object
        id = new IdentificationImpl(nm.elementAt(0), "::"+nm.elementAt(i),"");
        rt = ((SContainer) con).slookup_name(id);
        if (rt.length==0) throw new TIRExceptCommit("Can't find (1) object " + nm.name());
        if (rt.length==1) {
          if (!(rt[0] instanceof Container))
            throw new TIRExceptCommit("Can't find (2) object " + nm.name());
          con = (Container) rt[0];
        } else {  // type in struct (etc.) - more versions of this struct
          boolean found = false;
          id.version(akt.obj.version());  // substruct version is inherited from parent
          int j;
          for (j=0;j<rt.length;j++) {
            if (id.is_short_equal(rt[j].get_identification())) {
              found = true;
              break;  // -- break for
	    }
	  } // -- end of for
	  if (!found) throw new TIRExceptCommit("Can't find (3) object " + nm.name());
          if (!(rt[j] instanceof Container))
            throw new TIRExceptCommit("Can't find (4) object " + nm.name());
	  con = (Container) rt[j];
	}
      }
      if ((ret = ((SContainer)con).slookup(akt.obj))==null)
        throw new TIRExceptCommit("Can't find (5) object " + nm.name());
        
      akt = akt.next;
    }
    } catch (TIRExceptLock e) {}
  }
  
  public void doCommit(Container in, Repository rep) throws RemoteException {
    if (normal!=null)
      normal.contentId = ((WorkRepositoryImpl) rep).workId;
    if (isNewP) {
      stKindImpl.toNormal();
      this.rep = rep;
    }
  }
  
  public void doAbort(long workId) throws RemoteException {}
}
