/* $Id: GenSpecFile.java,v 1.1 2003/01/15 22:38:25 hnetynka Exp $ */
package SOFA.SOFAnode.Run;
import java.io.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import SOFA.SOFAnode.Made.TIR.*;
import SOFA.SOFAnode.Made.TIR.Access.TIRAccessMethods;

/** Generates SOFA specfile for the choosen application.
  * 
  * @author Petr Hnetynka
  */
public class GenSpecFile {

  /** Main method. Neccessary properies: sofa.rmihost, sofa.rmiport, sofa.tr.url.
    * 
    * @param argv 1st argument - full application name, 2nd name of the generated specfile
    * 
    */
  public static void main(String[] argv) {
    if (argv.length < 2) {
      System.out.println("Too few arguments.");
      System.out.println("  1st argument - full application name");
      System.out.println("  2nd argument - name of the generated specfile");
      System.exit(1);
    }

    if (!argv[0].matches("^.+\\[.+\\]$")) {
      System.out.println("Bad format of application name.");
      System.exit(1);
    }

    Document depldesc = null;
    try {
      depldesc = TRAccess.getDeplDescrXMLLocal(argv[0], "index.dc");
    } catch (java.net.MalformedURLException e) {
      System.out.println("MalformedURLException: "+e.getMessage());
      e.printStackTrace();
      System.exit(1);
    } catch (java.io.IOException e) {
      System.out.println("IOException: "+e.getMessage());
      System.exit(1);
    } catch (javax.xml.parsers.ParserConfigurationException e) {
      System.out.println("ParserConfigurationExceptioni: "+e.getMessage() );
      System.exit(1);
    } catch (org.xml.sax.SAXException e) {
      System.out.println("SAXException: "+e.getMessage() );
      System.exit(1);
    }

    Element applRootElement = depldesc.getDocumentElement();
    applRootElement.normalize();

    if (applRootElement.getTagName().compareTo("sofa_system") != 0) {
      System.out.println(argv[0]+" is not system component.");
      System.exit(1);
    }

    Repository tir = null;
    try {
      tir = TIRAccessMethods.getRepository();
    } catch (java.net.MalformedURLException e) {
      System.out.println("MalformedURLException: "+e.getMessage() );
      System.exit(1);
    } catch (java.rmi.NotBoundException e) {
      System.out.println("NotBoundException: "+e.getMessage() );
      System.exit(1);
    } catch (java.rmi.RemoteException e) {
      System.out.println("RemoteException: "+e.getMessage() );
      System.exit(1);
    }
    
    System.out.print("Generating specfile...");
    
    applRootElement = processElement(tir, null, null, applRootElement);
    applRootElement = addApplName(applRootElement);

    
    SpecFile sf = SpecFile.getSpecFileFromDeplDescriptorXMLDoc(applRootElement);
    try {
      sf.toFile(argv[1]+".ssf");
    } catch  (TransformerConfigurationException e) {
      System.out.println("TransformerConfigurationException: "+e.getMessage() );
      System.exit(1);
    } catch (TransformerException e) {
      System.out.println("TransformerException: "+e.getMessage() );
      System.exit(1);
    } catch (java.io.IOException e) {
      System.out.println("IOException: "+e.getMessage());
      System.exit(1);
    }

    System.out.println("OK");
  }

  private static Element processElement(Repository tir, String inst, InfoArray riarray, Element element) {
    String archName = null;
    InfoArray iarray = null;
    NodeList nl = element.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if ((n.getNodeType()==Node.ELEMENT_NODE) && (n.getNodeName().compareTo("component_ref")==0)) {
        String instName = ((Element)n).getAttribute("inst");
        String name = ((Element)n).getAttribute("arch");
        String version = ((Element)n).getAttribute("version");
        Element archRootElement = null;
        try {
          Document archDoc = TRAccess.getDeplDescrXMLLocal(name+"["+version+"]", "index.dc"); 
          archRootElement = archDoc.getDocumentElement();
          archRootElement.normalize();
        } catch (java.net.MalformedURLException e) {
          System.out.println("MalformedURLException: "+e.getMessage());
          e.printStackTrace();
          System.exit(1);
        } catch (java.io.IOException e) {
          System.out.println("IOException: "+e.getMessage());
          System.exit(1);
        } catch (javax.xml.parsers.ParserConfigurationException e) {
          System.out.println("ParserConfigurationException: "+e.getMessage() );
          System.exit(1);
        } catch (org.xml.sax.SAXException e) {
          System.out.println("SAXException: "+e.getMessage() );
          System.exit(1);
        }
        archRootElement.setAttribute("name", instName);
        archRootElement = processElement(tir, instName, iarray, archRootElement);
        element.removeChild(n);
        element.appendChild(element.getOwnerDocument().importNode(archRootElement, true));
      } else if ((n.getNodeType()==Node.ELEMENT_NODE) && (n.getNodeName().compareTo("architecture_ref")==0)) {
        archName = n.getFirstChild().getNodeValue();
        iarray = getBindConnInfo(tir, archName);
      } else if ((n.getNodeType()==Node.ELEMENT_NODE) && (n.getNodeName().compareTo("connector")==0)) {
        String connName = ((Element)n).getAttribute("name");
        if (connName.startsWith("_delegate_") || connName.startsWith("_subsume_")) {
          ((Element) n).setAttribute("frame", "SGenerator:CSProcCallImpl");
        } else {
          String usingConn = riarray.getConn(inst, connName);
          if (usingConn == null) {
            System.out.println("Mismatch in TR and TIR.\nCannot find connector for "+inst+":"+connName+".");
            System.exit(1);
          } else {
            ((Element) n).setAttribute("frame", "SGenerator:"+usingConn+"Impl");
          }
        }
        Element newEl = n.getOwnerDocument().createElement("transport");
        newEl.setAttribute("type", "LOCAL");
        n.appendChild(newEl);
      }
    }
    
    return element;
  }

  private static Element addApplName(Element element) {
    NodeList nl = element.getChildNodes();
    for (int i=0; i<nl.getLength(); i++) {
      Node n = nl.item(i);
      if ((n.getNodeType()==Node.ELEMENT_NODE) && (n.getNodeName().compareTo("architecture_ref")==0)) {
        String ar = n.getFirstChild().getNodeValue();
        ar = ar.substring(ar.lastIndexOf(':')+1, ar.indexOf('?'));
        element.setAttribute("name", ar);
      }
    }
    return element;
  }

  public static class BindConnInfo {
    String inst;
    String prname;
    String con;
    public int type;
    public static final int PROV = 0;
    public static final int REQ = 1;
    public BindConnInfo(String i, String pr, String c, int t) {
      inst = i;
      prname = pr;
      con = c;
      type = t;
    }
  }

  public static class InfoArray extends java.util.ArrayList {
    public void addInfo(String inst, String pr, String using, int type) {
      add(new BindConnInfo(inst, pr, using, type));
    }
  
    public BindConnInfo getInfo(String inst, String pr) {
      BindConnInfo bi = null;
      for (int i=0; i<size(); i++) {
        bi = (BindConnInfo) get(i);
        if ((bi.inst.compareTo(inst) == 0) && (bi.prname.compareTo(pr) == 0)) {
          return bi;
        }
      }
      return null;
    }

    public String getConn(String inst, String pr) {
      BindConnInfo i = getInfo(inst, pr);
      if (i != null)
        return i.con;
      else
        return null;
    }
  }

  public static InfoArray getBindConnInfo(Repository rep, String archFullName) {
    try {
    ArchitectureDef arch = (ArchitectureDef) TIRAccessMethods.lookupCDLContained(rep, archFullName);
      if (arch != null) {
        InfoArray iarray = new InfoArray();
        Contained[] binds = arch.contents(rep.get_spec_def_kind(DefinitionKind.dk_Bind));
        for (int i=0; i<binds.length; i++) {
          BindDef bind = (BindDef) binds[i];
          String using = bind.using();
          switch (bind.mode().value()) {
            case BindMode.DELEGATE:
              {
                BindOperSub rhs = (BindOperSub) bind.rhsop();
                BindType[] relems = rhs.elements();
                BindType[] rsubs = rhs.subs();
                String rcompname;
                if (rsubs[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                  rcompname = ((BindTypeNormal) rsubs[0]).name();
                } else {
                  rcompname = ((BindTypeArray) rsubs[0]).name();
                }
                String relemname;
                if (relems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                  relemname = ((BindTypeNormal) relems[0]).name();
                } else {
                  relemname = ((BindTypeArray) relems[0]).name();
                }
                iarray.addInfo(rcompname, relemname, using, BindConnInfo.PROV);
              }
              break;
            case BindMode.SUBSUME:
              {
                BindOperSub lhs = (BindOperSub) bind.lhsop();
                BindType[] lelems = lhs.elements();
                BindType[] lsubs = lhs.subs();
                String lcompname;
                if (lsubs[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                  lcompname = ((BindTypeNormal) lsubs[0]).name();
                } else {
                  lcompname = ((BindTypeArray) lsubs[0]).name();
                }
                String lelemname;
                if (lelems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                  lelemname = ((BindTypeNormal) lelems[0]).name();
                } else {
                  lelemname = ((BindTypeArray) lelems[0]).name();
                }
                iarray.addInfo(lcompname, lelemname, using, BindConnInfo.REQ);
              }
              break;
            case BindMode.BIND:
              if (bind.lhsop().get_bo_kind().value() == BindOperKind.bok_provreq) {
                ;
              } else {
                BindOperSub lhs = (BindOperSub) bind.lhsop();
                BindOperSub rhs = (BindOperSub) bind.rhsop();
                BindType[] lelems = lhs.elements();
                BindType[] lsubs = lhs.subs();
                BindType[] relems = rhs.elements();
                BindType[] rsubs = rhs.subs();
                String lcompname;
                String rcompname;
                if (lsubs[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                  lcompname = ((BindTypeNormal) lsubs[0]).name();
                } else {
                  lcompname = ((BindTypeArray) lsubs[0]).name();
                }
                if (rsubs[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                  rcompname = ((BindTypeNormal) rsubs[0]).name();
                } else {
                  rcompname = ((BindTypeArray) rsubs[0]).name();
                }
                String lelemname;
                String relemname;
                if (lelems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                  lelemname = ((BindTypeNormal) lelems[0]).name();
                } else {
                  lelemname = ((BindTypeArray) lelems[0]).name();
                }
                if (relems[0].get_bt_kind().value() == BindTypeKind.btk_normal) {
                  relemname = ((BindTypeNormal) relems[0]).name();
                } else {
                  relemname = ((BindTypeArray) relems[0]).name();
                }
                iarray.addInfo(lcompname, lelemname, using, BindConnInfo.REQ);
                iarray.addInfo(rcompname, relemname, using, BindConnInfo.PROV);
              }
              break;
          }
        }
        return iarray;
      } else {
        return null;
      }
    } catch (java.rmi.RemoteException e) {
      System.out.println("RemoteException: "+e.getMessage());
      e.printStackTrace();
      System.exit(1);
    } catch (TIRExceptLock e) {
      System.out.println("TIRExceptLock: "+e.getMessage());
      e.printStackTrace();
      System.exit(1);
    }
    return null;
  }

}
