/* $Id: Launcher.java,v 1.3 2002/07/31 10:15:39 hnetynka Exp $ */
package SOFA.SOFAnode.Run;

import SOFA.SOFAnode.Run.DeplDockRegistry.*;
import SOFA.SOFAnode.Run.Deployment.DeplDock;

import org.omg.CosNaming.*;

/** Launcher of SOFA applications
  *
  * @author Petr Hnetynka
  */
public class Launcher {
 
  /** Launching of SOFA applications. Depl. descriptor must be set as a 1st cmdline argument,
    * 2nd argument is name of the home depldock. Set also all necessary properties
    * ("sofa.sofanode", "sofa.rmihost", "sofa.rmiport","sofa.orbport", "sofa.orbhost".
    * All depldocks necessary for application must run.
    */
  public static void main(String argv[]) {
    if (argv.length != 2) {
      System.out.println("Bad number of arguments.");
      System.out.println("Specify file with DD and name of the home deployment dock");
      System.exit(1);
    }

    if (System.getSecurityManager() == null) {
      System.setSecurityManager(new java.rmi.RMISecurityManager());
    }

    SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor depldesc = null;
    try {
      System.out.print("Parsing deployment descriptor...");
      depldesc = new SOFA.SOFAnode.Run.Deployment.DeploymentDescriptorImpl(argv[0]);
      System.out.println("OK");
    } catch (java.io.IOException e) {
      System.out.println("Can't opent deployment descriptor");
      System.exit(1);
    } catch (javax.xml.parsers.ParserConfigurationException e) {
      System.out.println("ParserConfigurationExceptioni: "+e.getMessage() );
      System.exit(1);
    } catch (org.xml.sax.SAXException e) {
      System.out.println("SAXException: "+e.getMessage() );
      System.exit(1);
    }

    SOFA.Util.CORBAAccess ca = new SOFA.Util.CORBAAccess();

    NameComponent nc = new NameComponent("DeplDockRegistry", "");
    NameComponent path[] = {nc};
    DeplDockRegistry ddr = null;
    try {
      System.out.print("Obtaining depldock registry...");
      ddr = DeplDockRegistryHelper.narrow(SOFA.Util.CORBAAccess.ncRef.resolve(path));
      System.out.println("OK");
    } catch (Exception e) {
      System.out.println("Can't get DeplDock registry.");
      e.printStackTrace();
      System.exit(1);
    }
    
    
    byte[] serRef = null;
    try {
      System.out.print("Obtaining reference to depldock...");
      serRef = ddr.lookup(argv[1]);
      System.out.println("OK");
    } catch (NotFoundException e) {
      System.out.println("No deployment dock with name \""+argv[1]+"\".");
      System.exit(1);
    }
    
    SOFA.Connector.Reference ddRef = SOFA.SOFAnode.Run.Deployment.DeplDockImpl.bytesToSofaReference(serRef);
   
    DeplDock ddock = null;
    try {
      System.out.print("Creating connector to depldock...");
      ddock = (DeplDock) SOFA.Connector.Boot.DeplDockConnector.createClt(ddRef);
      System.out.println("OK");
    } catch (SOFA.Connector.ConnectorException e) {
      System.out.println("Can't create connector to deployment dock.");
      System.exit(1);
    }
    
    SOFA.Connector.Reference apl = null;
    try {
      System.out.print("Instatiate component...");
      apl = ddock.instantiate(depldesc, "");
      System.out.println("OK");
    } catch (SOFA.SOFAnode.Run.Deployment.DeploymentException e) {
      System.out.println("DeploymentException: "+e.getMessage());
      e.printStackTrace();
      System.exit(1);
    }

    SOFA.Component.DCUP.DCUPComponentManager aplcm = null;
    try {
      System.out.print("Creating connector to aplication component manager...");
      aplcm = (SOFA.Component.DCUP.DCUPComponentManager) SOFA.Connector.Boot.DCUPComponentManagerConnector.createClt(apl);
      System.out.println("OK");
    } catch (SOFA.Connector.ConnectorException e) {
      System.out.println("Can't create connector to aplication component manager.");
      System.exit(1);
    }
    
    try {
      System.out.print("Launching component...");
      aplcm.createComponent(new StorageImpl());
      System.out.println("OK");
    } catch (SOFA.Component.ComponentLifecycleException e) {
      System.out.println("ComponentLifecycleException: "+e.getMessage());
      e.printStackTrace();
      System.exit(1);
    }

    System.out.println(apl);
    System.out.println("Application is runnig :-)");
  }
}
