/* $Id: SOFABaseClassLoader.java,v 1.4 2002/10/08 16:14:15 hnetynka Exp $ */
package SOFA.SOFAnode.Run;

import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

/** SOFA base classloader. It loads implementation classes of applications, i.e. application class loader.
  *
  * @author Petr Hnetynka
  */
public class SOFABaseClassLoader  extends ClassLoader {
  
  /** URL with classes */
  URL url;
  /** Version of component implementation.  */
  String version;
  /** Full name of the component's architecture */
  String component;
  /** Path to the component classes */
  String pathComponent;

  /** Constructor.
    * @param parent    parent classloader
    * @param _url      the URL from which to load classes
    * @param _component full name of the component's architecture (e.g. ::CUNI::SOFA::demos::logdemo::logdemo)
    * @param _version  implementation version of classes
    */
  public SOFABaseClassLoader(ClassLoader parent, URL _url, String _component, String _version) {
    super(parent);
    url = _url;
    version = _version;
    component = _component;
    java.util.StringTokenizer tok = new java.util.StringTokenizer(component, "::");
    StringBuffer path = new StringBuffer("impl/");
    String nm = "";
    while (tok.hasMoreTokens()) {
      nm = tok.nextToken();
      path.append(nm);
      path.append("/");
    }
    pathComponent = path.toString();
//    System.out.println("URL: "+url);
//    System.out.println("version: "+version);
  }
  
  /** Set implementation version of classes.*/
  public void setVersion(String _version) {
    version = _version;
  }

  public Class findClass(String name) throws ClassNotFoundException {
    byte[] b;
//    System.out.println("???: "+name);
    b = loadClassData(name);
    Class ret = defineClass(name, b, 0, b.length);
//    System.out.println("??? End: "+name);
    return ret;
  }

  private byte[] loadClassData(String name) throws ClassNotFoundException {
//    System.out.println("loadClassData name = "+name);

    java.util.StringTokenizer tok = new java.util.StringTokenizer(name, ".");
    StringBuffer path = new StringBuffer(url.toString()+pathComponent+version);
    String nm = "";
    while (tok.hasMoreTokens()) {
      nm = tok.nextToken();
      path.append("/");
      path.append(nm);
    }
    path.append(".class");
    
    URL defURL;
    try {
      defURL = new URL(path.toString());
    } catch (MalformedURLException e) {
      throw new ClassNotFoundException("MalformedURLException: "+e.getMessage());
    }
    try {
      URLConnection con = defURL.openConnection();
      InputStream is = con.getInputStream();
      ByteArrayOutputStream bo = new ByteArrayOutputStream();
      int a = is.read();
      while (a != -1) {
        bo.write(a);
        a = is.read();
      }
      is.close();
      byte[] ar = bo.toByteArray();
      bo.close();
      return ar;
    } catch (IOException e) {
      throw new ClassNotFoundException("IOException: "+e.getMessage());
    }
  }

  /** Returns sofa class loader for specified version of implementation classes.
    * Base url (from which to load classes) is specified by java property <tt>sofa.tr.url</tt>
    * @param parent    parent classloader - it MUST be SOFA interface loader
    * @param component absolute name of component's architecture
    * @param version   version of implementation
    */
  public static ClassLoader getBaseClassLoader(ClassLoader parent, String component, String version) throws SOFARuntimeException {
    try {
      String baseURL = System.getProperty("sofa.tr.url","file:/sofa/tr/");
      return new SOFABaseClassLoader(parent, new URL(baseURL), component, version);
    } catch (MalformedURLException e) {
      throw new SOFARuntimeException("MalformedURLException: "+e.getMessage());
    }
  }
}
