/* $Id: TRAccess.java,v 1.1 2003/01/15 22:38:25 hnetynka Exp $ */
package SOFA.SOFAnode.Run;
import java.io.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import org.xml.sax.*;

/** Methods for simple access to TR.
  * 
  * @author Petr Hnetynka
  */
public class TRAccess {

  /** Returns depldescr of the given application from TR (it can access TR remotely).
    * It use property sofa.tr.url.
    * 
    * @param applname  full application name
    * @param ddname    name of the descriptor
    * @return   deplyment descriptor
    * @throws java.net.MalformedURLException  malformed url
    * @throws java.io.IOException             io exception
    * @throws javax.xml.parsers.ParserConfigurationException  xml exception
    * @throws org.xml.sax.SAXException                        xml exception
    */
  public static SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor getDeplDescr(String applname, String ddname) throws java.net.MalformedURLException, java.io.IOException, javax.xml.parsers.ParserConfigurationException, org.xml.sax.SAXException {
    String trString = System.getProperty("sofa.tr.url","file:/sofa/tr/");
    String path = trString+"impl";
    path = path + fullNameToURLName(applname, ddname);
    java.net.URL url = new java.net.URL(path);
    java.net.URLConnection con = url.openConnection();
    java.io.InputStream is = con.getInputStream();
    SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor dd = new SOFA.SOFAnode.Run.Deployment.DeploymentDescriptorImpl(is);
    is.close();
    return dd;
  }

  /** Returns depldescr of the given application from TR (it can access TR directly).
    * It use property sofa.tr.dir.
    * 
    * @param applname  full application name
    * @param ddname    name of the descriptor
    * @return   deplyment descriptor
    * @throws java.io.IOException             io exception
    * @throws javax.xml.parsers.ParserConfigurationException  xml exception
    * @throws org.xml.sax.SAXException                        xml exception
    */
  public static SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor getDeplDescrLocal(String applname, String ddname) throws java.net.MalformedURLException, java.io.IOException, javax.xml.parsers.ParserConfigurationException, org.xml.sax.SAXException {
    String trString = System.getProperty("sofa.tr.dir","/sofa/tr");
    String path = trString+File.separator+"impl";
    path = path + fullNameToURLName(applname, ddname);
    java.io.FileInputStream is = new FileInputStream(path);
    SOFA.SOFAnode.Run.Deployment.DeploymentDescriptor dd = new SOFA.SOFAnode.Run.Deployment.DeploymentDescriptorImpl(is);
    is.close();
    return dd;
  }

  /** Returns depldescr as an XML document of the given application (it can access TR remotely).
    * It use property sofa.tr.url.
    * 
    * @param applname  full application name
    * @param ddname    name of the descriptor
    * @return   XML document
    * @throws java.net.MalformedURLException  malformed url
    * @throws java.io.IOException             io exception
    * @throws javax.xml.parsers.ParserConfigurationException  xml exception
    * @throws org.xml.sax.SAXException                        xml exception
    */
  public static Document getDeplDescrXML(String applname, String ddname) throws java.net.MalformedURLException, java.io.IOException, javax.xml.parsers.ParserConfigurationException, org.xml.sax.SAXException {
    String trString = System.getProperty("sofa.tr.url","file:/sofa/tr/");
    String path = trString+"impl";
    path = path + fullNameToURLName(applname, ddname);
    java.net.URL url = new java.net.URL(path);
    java.net.URLConnection con = url.openConnection();
    java.io.InputStream is = con.getInputStream();
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder builder = factory.newDocumentBuilder();
    Document document = builder.parse(is);
    is.close();
    return document;
  }

  /** Returns depldescr as an XML document of the given application (it can access TR directly).
    * It use property sofa.tr.dir.
    * 
    * @param applname  full application name
    * @param ddname    name of the descriptor
    * @return   XML document
    * @throws java.io.IOException             io exception
    * @throws javax.xml.parsers.ParserConfigurationException  xml exception
    * @throws org.xml.sax.SAXException                        xml exception
    */
  public static Document getDeplDescrXMLLocal(String applname, String ddname) throws java.net.MalformedURLException, java.io.IOException, javax.xml.parsers.ParserConfigurationException, org.xml.sax.SAXException {
    String trString = System.getProperty("sofa.tr.dir","/sofa/tr");
    String path = trString+File.separator+"impl";
    path = path + fullNameToURLName(applname, ddname);
    java.io.FileInputStream is = new FileInputStream(path);
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder builder = factory.newDocumentBuilder();
    Document document = builder.parse(is);
    is.close();
    return document;
  }

  /** Converts full application name to the URL path and append appendName.
    * 
    * @param clname      full application name
    * @param appendName  append this string to path
    * @return            URL path ("::" converted to "/")
    */
  public static String fullNameToURLName(String clname, String appendName) {
    clname = clname.replaceAll("::", "/");
    clname = clname.replace('[', '/');
    clname = clname.replace(']', '/');
    clname = clname + appendName;
    if (clname.charAt(0) != '/')
      clname = "/" + clname;
    return clname;
  }
  
}
